from __future__ import annotations
import inspect
from typing import Callable, Union, get_args, get_origin
import typing

# In Python 3.10+, the `|` operator creates a `types.UnionType`.
# We need to handle this for forward compatibility, but it won't exist in 3.9.
try:
    from types import UnionType

    UNION_TYPES = (Union, UnionType)
except ImportError:
    UNION_TYPES = (Union,)

from fpy.error import BackendError
from fpy.ir import Ir, IrGoto, IrIf, IrLabel, IrPushLabelOffset
from fpy.model import DirectiveErrorCode
from fpy.types import (
    MAX_DIRECTIVES_COUNT,
    SIGNED_INTEGER_TYPES,
    SPECIFIC_NUMERIC_TYPES,
    UNSIGNED_INTEGER_TYPES,
    CompileState,
    Emitter,
    FieldSymbol,
    FppType,
    CastSymbol,
    CommandSymbol,
    FpyFloatValue,
    FunctionSymbol,
    BuiltinSymbol,
    TypeCtorSymbol,
    VariableSymbol,
    FpyIntegerValue,
    FpyStringValue,
    NothingValue,
    Visitor,
    is_instance_compat,
)

from fpy.bytecode.directives import (
    BINARY_STACK_OPS,
    UNARY_STACK_OPS,
    AllocateDirective,
    ArrayIndexType,
    BinaryStackOp,
    CallDirective,
    ConstCmdDirective,
    DiscardDirective,
    ExitDirective,
    FloatDivideDirective,
    FloatExtendDirective,
    FloatToSignedIntDirective,
    FloatToUnsignedIntDirective,
    FloatTruncateDirective,
    FwOpcodeType,
    GotoDirective,
    IfDirective,
    IntegerSignedExtend16To64Directive,
    IntegerSignedExtend32To64Directive,
    IntegerSignedExtend8To64Directive,
    IntegerTruncate64To16Directive,
    IntegerTruncate64To8Directive,
    IntegerZeroExtend16To64Directive,
    IntegerZeroExtend32To64Directive,
    IntegerZeroExtend8To64Directive,
    PeekDirective,
    FloatMultiplyDirective,
    GetFieldDirective,
    IntAddDirective,
    IntMultiplyDirective,
    LoadRelDirective,
    LoadAbsDirective,
    MemCompareDirective,
    NoOpDirective,
    IntegerTruncate64To32Directive,
    ReturnDirective,
    SignedGreaterThanOrEqualDirective,
    SignedIntToFloatDirective,
    StackCmdDirective,
    Directive,
    NotDirective,
    PushValDirective,
    StackSizeType,
    StoreRelConstOffsetDirective,
    StoreAbsConstOffsetDirective,
    StoreRelDirective,
    StoreAbsDirective,
    PushPrmDirective,
    PushTlmValDirective,
    UnaryStackOp,
    UnsignedIntToFloatDirective,
)
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.templates.prm_template import PrmTemplate
from fprime_gds.common.models.serialize.array_type import ArrayType as ArrayValue
from fprime_gds.common.models.serialize.bool_type import BoolType as BoolValue
from fprime_gds.common.models.serialize.numerical_types import (
    U8Type as U8Value,
    U64Type as U64Value,
    I32Type as I32Value,
    I64Type as I64Value,
    F32Type as F32Value,
    F64Type as F64Value,
    IntegerType as IntegerValue,
)
from fpy.syntax import (
    Ast,
    AstAssert,
    AstBinaryOp,
    AstStmtList,
    AstBreak,
    AstContinue,
    AstDef,
    AstExpr,
    AstFor,
    AstMemberAccess,
    AstIndexExpr,
    AstLiteral,
    AstNodeWithSideEffects,
    AstReturn,
    AstBlock,
    AstBlock,
    AstIf,
    AstAssign,
    AstFuncCall,
    AstUnaryOp,
    AstVar,
    AstWhile,
)


class AssignVariableOffsets(Visitor):
    """Assigns frame offsets to variables before code generation.

    This must run before GenerateFunctions so that global variable offsets
    are known when generating function bodies that access them.
    """

    def visit_AstBlock(self, node: AstBlock, state: CompileState):
        # Assign offsets to variables in this scope
        lvar_array_size_bytes = 0
        for name, sym in state.local_scopes[node].items():
            if not is_instance_compat(sym, VariableSymbol):
                # doesn't require space to be allocated
                continue
            if sym.frame_offset is not None:
                # Already has an offset (e.g., function argument or global)
                continue
            # Assign new offset
            sym.frame_offset = lvar_array_size_bytes
            lvar_array_size_bytes += sym.type.getMaxSize()

    def visit_AstDef(self, node: AstDef, state: CompileState):
        # Assign offsets for function arguments (negative offsets before frame start)
        # The args come before the call frame header (return addr + prev frame ptr)
        # so they have negative offsets relative to the function's stack frame start
        func = state.resolved_symbols[node.name]
        if func.args is None or len(func.args) == 0:
            # no args
            return

        # Args are pushed left-to-right, so first arg is deepest
        # After CALL, the stack looks like:
        #   [args...] [return_addr] [prev_frame_ptr] <-- stack_frame_start points here
        # So to access args, we use negative offsets from stack_frame_start
        from fpy.model import STACK_FRAME_HEADER_SIZE

        arg_offset = -STACK_FRAME_HEADER_SIZE
        for arg in reversed(func.args):
            arg_name, arg_type, _ = arg
            arg_var = state.local_scopes[node.body][arg_name]
            arg_offset -= arg_type.getMaxSize()
            arg_var.frame_offset = arg_offset


class GenerateFunctionEntryPoints(Visitor):
    def visit_AstDef(self, node: AstDef, state: CompileState):
        entry_label = IrLabel(node, "entry")
        state.func_entry_labels[node] = entry_label


class GenerateFunctions(Visitor):
    def visit_AstDef(self, node: AstDef, state: CompileState):
        entry_label = state.func_entry_labels[node]
        code = [entry_label]
        code.extend(GenerateFunctionBody().emit(node.body, state))
        func = state.resolved_symbols[node.name]
        if func.return_type is NothingValue and not state.does_return[node.body]:
            # implicit empty return
            arg_bytes = sum(arg[1].getMaxSize() for arg in (func.args or []))
            code.append(ReturnDirective(0, arg_bytes))
        state.generated_funcs[node] = code


class GenerateFunctionBody(Emitter):
    # Flag indicating we're generating code inside a function body
    # This affects how we access global variables (need GLOBAL directives)
    in_function = True

    def try_emit_expr_as_const(
        self, node: AstExpr, state: CompileState
    ) -> Union[list[Directive | Ir], None]:
        """if the expr has a compile time const value, emit that as a PUSH_VAL"""
        expr_value = state.contextual_values.get(node)

        if expr_value is None:
            # no const value
            return None

        assert not is_instance_compat(
            expr_value, (FpyIntegerValue, FpyStringValue, FpyFloatValue)
        ), expr_value

        if is_instance_compat(expr_value, NothingValue):
            # nothing type has no value
            return []

        # it has a constant value at compile time
        serialized_expr_value = expr_value.serialize()

        # push it to the stack
        return [PushValDirective(serialized_expr_value)]

    def discard_expr_result(self, node: Ast, state: CompileState) -> list[Directive]:
        """if the node is an expr, generate code to discard its stack value"""
        if not is_instance_compat(node, AstExpr):
            # nothing to discard
            return []

        result_type = state.contextual_types[node]
        if result_type == NothingValue:
            return []
        if result_type.getMaxSize() > 0:
            return [DiscardDirective(result_type.getMaxSize())]
        return []

    def get_64_bit_numeric_type(self, type: FppType) -> FppType:
        """return the 64 bit version of the input numeric type"""
        assert type in SPECIFIC_NUMERIC_TYPES, type
        return (
            I64Value
            if type in SIGNED_INTEGER_TYPES
            else U64Value if type in UNSIGNED_INTEGER_TYPES else F64Value
        )

    def convert_numeric_type(
        self, from_type: FppType, to_type: FppType
    ) -> list[Directive]:
        """
        return a list of dirs needed to convert a numeric stack value of from_type to a stack value of to_type
        """
        if from_type == to_type:
            return []

        # only valid runtime type conversion is between two numeric types
        assert (
            from_type in SPECIFIC_NUMERIC_TYPES and to_type in SPECIFIC_NUMERIC_TYPES
        ), (
            from_type,
            to_type,
        )

        dirs = []
        # first go to 64 bit width
        dirs.extend(self.extend_numeric_type_to_64_bits(from_type))
        from_64_bit = self.get_64_bit_numeric_type(from_type)
        to_64_bit = self.get_64_bit_numeric_type(to_type)

        # now convert between int and float if necessary
        if from_64_bit == U64Value and to_64_bit == F64Value:
            dirs.append(UnsignedIntToFloatDirective())
            from_64_bit = F64Value
        elif from_64_bit == I64Value and to_64_bit == F64Value:
            dirs.append(SignedIntToFloatDirective())
            from_64_bit = F64Value
        elif from_64_bit == U64Value or from_64_bit == I64Value:
            assert to_64_bit == U64Value or to_64_bit == I64Value
            # conversion from signed to unsigned int is implicit, doesn't need code gen
            from_64_bit = to_64_bit
        elif from_64_bit == F64Value and to_64_bit == I64Value:
            dirs.append(FloatToSignedIntDirective())
            from_64_bit = I64Value
        elif from_64_bit == F64Value and to_64_bit == U64Value:
            dirs.append(FloatToUnsignedIntDirective())
            from_64_bit = U64Value

        assert from_64_bit == to_64_bit, (from_64_bit, to_64_bit)

        # now truncate back down to desired size
        dirs.extend(
            self.truncate_numeric_type_from_64_bits(to_64_bit, to_type.getMaxSize())
        )
        return dirs

    def truncate_numeric_type_from_64_bits(
        self, from_type: FppType, new_size: int
    ) -> list[Directive]:

        assert new_size in (1, 2, 4, 8), new_size
        assert from_type.getMaxSize() == 8, from_type.getMaxSize()

        if new_size == 8:
            # already correct size
            return []

        if from_type == F64Value:
            # only one option for float trunc
            assert new_size == 4, new_size
            return [FloatTruncateDirective()]

        # must be an int
        assert issubclass(from_type, IntegerValue), from_type

        if new_size == 1:
            return [IntegerTruncate64To8Directive()]
        elif new_size == 2:
            return [IntegerTruncate64To16Directive()]

        return [IntegerTruncate64To32Directive()]

    def extend_numeric_type_to_64_bits(self, type: FppType) -> list[Directive]:
        if type.getMaxSize() == 8:
            # already 8 bytes
            return []
        if type == F32Value:
            return [FloatExtendDirective()]

        # must be an int
        assert issubclass(type, IntegerValue), type

        from_size = type.getMaxSize()
        assert from_size in (1, 2, 4, 8), from_size

        if type in SIGNED_INTEGER_TYPES:
            if from_size == 1:
                return [IntegerSignedExtend8To64Directive()]
            elif from_size == 2:
                return [IntegerSignedExtend16To64Directive()]
            else:
                return [IntegerSignedExtend32To64Directive()]
        else:
            if from_size == 1:
                return [IntegerZeroExtend8To64Directive()]
            elif from_size == 2:
                return [IntegerZeroExtend16To64Directive()]
            else:
                return [IntegerZeroExtend32To64Directive()]

    def calc_lvar_offset_of_array_element(
        self, node: Ast, idx_expr: AstExpr, array_type: FppType, state: CompileState
    ) -> list[Directive | Ir]:
        """generates code to push to stack the U64 byte offset in the array for an array access, while performing an array oob
        check. idx_expr is the expression to calculate the index, and dest is the FieldSymbol containing info about the
        dest array"""
        dirs = []
        # let's push the offset of base lvar first, then
        # calculate the offset in base type, then add

        # push the index to the stack, do a bounds check,
        dirs.extend(self.emit(idx_expr, state))
        # okay now let's do an array oob check
        # we want to peek the index so we can consume it for the oob check
        # byte count
        dirs.append(
            PushValDirective(StackSizeType(ArrayIndexType.getMaxSize()).serialize())
        )
        # offset
        dirs.append(PushValDirective(StackSizeType(0).serialize()))
        dirs.append(PeekDirective())  # duplicate the index
        # convert idx to i64
        dirs.extend(self.convert_numeric_type(ArrayIndexType, I64Value))
        dirs.append(
            PushValDirective(I64Value(array_type.LENGTH).serialize())
        )  # push the length as I64
        # check if idx >= length
        dirs.append(SignedGreaterThanOrEqualDirective())
        # if true, fail with error code, otherwise go to after check
        oob_check_end_label = IrLabel(node, "oob_check_end")
        dirs.append(IrIf(oob_check_end_label))
        # push the error code we should fail with if false
        dirs.append(
            PushValDirective(
                U8Value(DirectiveErrorCode.ARRAY_OUT_OF_BOUNDS.value).serialize()
            )
        )
        dirs.append(ExitDirective())
        dirs.append(oob_check_end_label)
        # okay we're good. should still have the idx on the stack

        # multiply the index by the member type size
        dirs.append(PushValDirective(U64Value(array_type.MEMBER_TYPE.getMaxSize()).serialize()))
        dirs.append(IntMultiplyDirective())
        return dirs

    def emit_AstBlock(self, node: AstBlock, state: CompileState):
        dirs = []
        # Calculate lvar array size bytes (offsets already assigned by AssignVariableOffsets)
        lvar_array_size_bytes = 0
        for name, sym in state.local_scopes[node].items():
            if not is_instance_compat(sym, VariableSymbol):
                # doesn't require space to be allocated
                continue
            if sym.frame_offset < 0:
                # function argument (negative)
                continue
            if sym.is_global and self.in_function:
                # Global variable accessed from inside a function - already allocated at top level
                continue
            # Track the max offset to know how much space to allocate
            end_of_var = sym.frame_offset + sym.type.getMaxSize()
            if end_of_var > lvar_array_size_bytes:
                lvar_array_size_bytes = end_of_var

        bytes_to_allocate = lvar_array_size_bytes
        if bytes_to_allocate > 0:
            dirs.append(AllocateDirective(bytes_to_allocate))

        for stmt in node.stmts:
            if not is_instance_compat(stmt, AstNodeWithSideEffects):
                # if the stmt can't do anything on its own, ignore it
                # TODO warn
                continue
            dirs.extend(self.emit(stmt, state))
            # discard stack value if it was an expr
            dirs.extend(self.discard_expr_result(stmt, state))
        return dirs

    def emit_AstStmtList(self, node: AstStmtList, state: CompileState):
        dirs = []
        for stmt in node.stmts:
            if not is_instance_compat(stmt, AstNodeWithSideEffects):
                # if the stmt can't do anything on its own, ignore it
                # TODO warn
                continue
            dirs.extend(self.emit(stmt, state))
            # discard stack value if it was an expr
            dirs.extend(self.discard_expr_result(stmt, state))
        return dirs

    def emit_AstIf(self, node: AstIf, state: CompileState):
        dirs = []

        cases: list[tuple[AstExpr, AstStmtList]] = []

        cases.append((node.condition, node.body))

        for case in node.elifs:
            cases.append((case.condition, case.body))

        if_end_label = IrLabel(node, "end")

        for case in cases:
            case_end_label = IrLabel(case[1], "end")
            case_dirs = []
            # put the conditional on top of stack
            case_dirs.extend(self.emit(case[0], state))
            # include if stmt (update the end idx later)
            if_dir = IrIf(case_end_label)

            case_dirs.append(if_dir)
            # include body
            case_dirs.extend(self.emit(case[1], state))
            # once we've finished executing the body:
            # include a goto end of if
            case_dirs.append(IrGoto(if_end_label))
            case_dirs.append(case_end_label)

            dirs.extend(case_dirs)

        if node.els is not None:
            dirs.extend(self.emit(node.els, state))

        dirs.append(if_end_label)

        return dirs

    def emit_AstWhile(self, node: AstWhile, state: CompileState):
        # start by creating labels. store them in dicts so that break/continue
        # can use them
        while_start_label = IrLabel(node, "start")
        while_end_label = IrLabel(node, "end")
        for_loop_increment_label = None
        state.while_loop_start_labels[node] = while_start_label
        state.while_loop_end_labels[node] = while_end_label
        # if this used to be a for loop:
        if node in state.desugared_for_loops:
            # there should be at least one stmt in a for loop's body (the inc stmt)
            for_loop_increment_label = IrLabel(node, "increment")
            state.for_loop_inc_labels[node] = for_loop_increment_label

        dirs = [while_start_label]
        # push the condition to the stack
        dirs.extend(self.emit(node.condition, state))
        # if the cond is true, fall thru, otherwise go to end
        dirs.append(IrIf(while_end_label))
        # run body

        for stmt_idx, stmt in enumerate(node.body.stmts):
            if not is_instance_compat(stmt, AstNodeWithSideEffects):
                # if the stmt can't do anything on its own, ignore it
                continue
            # we're going to manually emit the body's stmts instead
            # of just emitting the body, because A) it doesn't matter
            # and B) we need the index of the last statement in the body
            # if we're a for loop, because that's where the continue stmt
            # needs to go
            if (
                stmt_idx == len(node.body.stmts) - 1
                and for_loop_increment_label is not None
            ):
                # last stmt, it must be the inc stmt, add the label before it
                dirs.append(for_loop_increment_label)
            dirs.extend(self.emit(stmt, state))
            # discard stack value if it was an expr
            dirs.extend(self.discard_expr_result(stmt, state))
        # go back to condition check
        dirs.append(IrGoto(while_start_label))
        dirs.append(while_end_label)

        return dirs

    def emit_AstBreak(self, node: AstBreak, state: CompileState):
        enclosing_loop = state.enclosing_loops[node]
        loop_end = state.while_loop_end_labels[enclosing_loop]
        return [IrGoto(loop_end)]

    def emit_AstContinue(self, node: AstContinue, state: CompileState):
        enclosing_loop = state.enclosing_loops[node]
        if enclosing_loop in state.desugared_for_loops:
            loop_start = state.for_loop_inc_labels[enclosing_loop]
        else:
            loop_start = state.while_loop_end_labels[enclosing_loop]
        return [IrGoto(loop_start)]

    def emit_AstDef(self, node: AstDef, state: CompileState):
        # don't generate other functions, just do this one
        return []

    def emit_AstReturn(self, node: AstReturn, state: CompileState):
        enclosing_func = state.enclosing_funcs[node]
        enclosing_func = state.resolved_symbols[enclosing_func.name]
        func_args_size = sum(arg[1].getMaxSize() for arg in enclosing_func.args)

        if node.value is not None:
            dirs = self.emit(node.value, state)
            value_size = state.contextual_types[node.value].getMaxSize()
        else:
            dirs = []
            value_size = 0
        dirs.append(ReturnDirective(value_size, func_args_size))

        return dirs

    def emit_AstFor(self, node: AstFor, state: CompileState):
        # should have been desugared out
        assert False, node

    def emit_AstIndexExpr(self, node: AstIndexExpr, state: CompileState):
        const_dirs = self.try_emit_expr_as_const(node, state)
        if const_dirs is not None:
            return const_dirs
        sym = state.resolved_symbols[node]

        assert is_instance_compat(sym, FieldSymbol), sym

        # use the unconverted for this expr for now, because we haven't run conversion
        unconverted_type = state.synthesized_types[node]
        # however, for parent, use converted because conversion has been run
        parent_type = state.contextual_types[node.parent]

        assert issubclass(parent_type, ArrayValue)
        assert unconverted_type == parent_type.MEMBER_TYPE, (
            parent_type.MEMBER_TYPE,
            unconverted_type,
        )

        # okay, we want to get an element from an array on the stack

        # TODO optimization: leave it in the lvar array instead of pushing the whole thing to stack
        # for now we push the whole thing
        dirs = self.emit(node.parent, state)

        # calculate the offset in the parent array
        dirs.extend(
            self.calc_lvar_offset_of_array_element(node, node.item, parent_type, state)
        )
        # truncate back to StackSizeType which is what get field uses
        dirs.extend(self.convert_numeric_type(U64Value, StackSizeType))

        # get the member from the stack at this offset, discard the rest of
        # the parent
        dirs.append(
            GetFieldDirective(
                parent_type.getMaxSize(), parent_type.MEMBER_TYPE.getMaxSize()
            )
        )

        # now convert the type if necessary
        converted_type = state.contextual_types[node]
        if unconverted_type != converted_type:
            dirs.extend(self.convert_numeric_type(unconverted_type, converted_type))

        return dirs

    def emit_AstVar(self, node: AstVar, state: CompileState):
        const_dirs = self.try_emit_expr_as_const(node, state)
        if const_dirs is not None:
            return const_dirs

        sym = state.resolved_symbols.get(node)

        assert is_instance_compat(sym, VariableSymbol), sym

        # Use global directives only when inside a function AND accessing a global variable
        # At top level, stack_frame_start = 0, so local and global offsets are the same
        use_global = self.in_function and sym.is_global
        if use_global:
            dirs = [LoadAbsDirective(sym.frame_offset, sym.type.getMaxSize())]
        else:
            dirs = [LoadRelDirective(sym.frame_offset, sym.type.getMaxSize())]

        unconverted_type = state.synthesized_types[node]
        converted_type = state.contextual_types[node]
        if unconverted_type != converted_type:
            dirs.extend(self.convert_numeric_type(unconverted_type, converted_type))

        return dirs

    def emit_AstMemberAccess(self, node: AstMemberAccess, state: CompileState):
        const_dirs = self.try_emit_expr_as_const(node, state)
        if const_dirs is not None:
            return const_dirs

        sym = state.resolved_symbols.get(node)

        if is_instance_compat(sym, dict):
            # don't generate code for it, it's a reference to a scope and
            # doesn't have a value
            return []

        # start with the unconverted type, because we haven't applied runtime type conversion yet
        unconverted_type = state.synthesized_types[node]

        dirs = []

        if is_instance_compat(sym, ChTemplate):
            dirs.append(PushTlmValDirective(sym.get_id()))
        elif is_instance_compat(sym, PrmTemplate):
            dirs.append(PushPrmDirective(sym.get_id()))
        elif is_instance_compat(sym, VariableSymbol):
            # Use global directives only when inside a function AND accessing a global variable
            use_global = self.in_function and sym.is_global
            if use_global:
                dirs.append(
                    LoadAbsDirective(sym.frame_offset, sym.type.getMaxSize())
                )
            else:
                dirs.append(LoadRelDirective(sym.frame_offset, sym.type.getMaxSize()))
        elif is_instance_compat(sym, FieldSymbol):
            # okay, put parent dirs in first
            dirs.extend(self.emit(sym.parent_expr, state))
            assert sym.local_offset is not None
            # use the converted type of parent
            parent_type = state.contextual_types[sym.parent_expr]
            # push the offset to the stack
            dirs.append(PushValDirective(StackSizeType(sym.local_offset).serialize()))
            dirs.append(
                GetFieldDirective(
                    parent_type.getMaxSize(), unconverted_type.getMaxSize()
                )
            )
        else:
            assert (
                False
            ), sym  # sym should either be impossible to put on stack or should have a compile time val

        converted_type = state.contextual_types[node]
        if converted_type != unconverted_type:
            dirs.extend(self.convert_numeric_type(unconverted_type, converted_type))

        return dirs

    def emit_AstBinaryOp(self, node: AstBinaryOp, state: CompileState):
        const_dirs = self.try_emit_expr_as_const(node, state)
        if const_dirs is not None:
            return const_dirs

        if node.op in (BinaryStackOp.AND, BinaryStackOp.OR):
            dirs = self.generate_short_circuit_boolean(node, state)
        else:
            # push lhs and rhs to stack
            dirs = self.emit(node.lhs, state)
            dirs.extend(self.emit(node.rhs, state))

            intermediate_type = state.op_intermediate_types[node]

            if (
                node.op == BinaryStackOp.EQUAL or node.op == BinaryStackOp.NOT_EQUAL
            ) and intermediate_type not in SPECIFIC_NUMERIC_TYPES:
                lhs_type = state.contextual_types[node.lhs]
                rhs_type = state.contextual_types[node.rhs]
                assert lhs_type == rhs_type, (lhs_type, rhs_type)
                dirs.append(MemCompareDirective(lhs_type.getMaxSize()))
                if node.op == BinaryStackOp.NOT_EQUAL:
                    dirs.append(NotDirective())
            elif (
                node.op == BinaryStackOp.FLOOR_DIVIDE and intermediate_type == F64Value
            ):
                # for float floor division, do float division, then convert to int, then
                # back to float
                dirs.append(FloatDivideDirective())
                dirs.append(FloatToSignedIntDirective())
                dirs.append(SignedIntToFloatDirective())
            else:

                dir = BINARY_STACK_OPS[node.op][intermediate_type]
                if dir != NoOpDirective:
                    # don't include no op
                    dirs.append(dir())

        # and convert the result of the op into the desired result of this expr
        unconverted_type = state.synthesized_types[node]
        converted_type = state.contextual_types[node]
        if unconverted_type != converted_type:
            dirs.extend(self.convert_numeric_type(unconverted_type, converted_type))

        return dirs

    def generate_short_circuit_boolean(
        self, node: AstBinaryOp, state: CompileState
    ) -> list[Directive | Ir]:
        dirs: list[Directive | Ir] = []
        end_label = IrLabel(node, "bool_end")

        if node.op == BinaryStackOp.AND:
            short_label = IrLabel(node, "and_short")
            dirs.extend(self.emit(node.lhs, state))
            # jump to short circuit when lhs is false
            dirs.append(IrIf(short_label))
            dirs.extend(self.emit(node.rhs, state))
            dirs.append(IrGoto(end_label))
            dirs.append(short_label)
            dirs.append(PushValDirective(BoolValue(False).serialize()))
        else:
            rhs_label = IrLabel(node, "or_rhs")
            dirs.extend(self.emit(node.lhs, state))
            # only evaluate rhs if lhs is false
            dirs.append(IrIf(rhs_label))
            dirs.append(PushValDirective(BoolValue(True).serialize()))
            dirs.append(IrGoto(end_label))
            dirs.append(rhs_label)
            dirs.extend(self.emit(node.rhs, state))

        dirs.append(end_label)
        return dirs

    def emit_AstUnaryOp(self, node: AstUnaryOp, state: CompileState):
        const_dirs = self.try_emit_expr_as_const(node, state)
        if const_dirs is not None:
            return const_dirs

        # push val to stack
        dirs = self.emit(node.val, state)

        # generate the actual op itself
        # which dir should we use?
        intermediate_type = state.op_intermediate_types[node]
        dir = UNARY_STACK_OPS[node.op][intermediate_type]

        if node.op == UnaryStackOp.NEGATE:
            # in this case, we also need to push -1
            if dir == FloatMultiplyDirective:
                dirs.append(PushValDirective(F64Value(-1).serialize()))
            elif dir == IntMultiplyDirective:
                dirs.append(PushValDirective(I64Value(-1).serialize()))

        dirs.append(dir())

        # and convert the result of the op into the desired result of this expr
        unconverted_type = state.synthesized_types[node]
        converted_type = state.contextual_types[node]
        if unconverted_type != converted_type:
            dirs.extend(self.convert_numeric_type(unconverted_type, converted_type))

        return dirs

    def emit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        const_dirs = self.try_emit_expr_as_const(node, state)
        if const_dirs is not None:
            return const_dirs

        node_args = node.args if node.args is not None else []
        func = state.resolved_symbols[node.func]
        dirs = []
        if is_instance_compat(func, CommandSymbol):
            const_args = not any(
                state.contextual_values[arg_node] is None for arg_node in node_args
            )
            if const_args:
                # can just hardcode this cmd
                arg_bytes = bytes()
                for arg_node in node_args:
                    arg_value = state.contextual_values[arg_node]
                    arg_bytes += arg_value.serialize()
                dirs.append(ConstCmdDirective(func.cmd.get_op_code(), arg_bytes))
            else:
                arg_byte_count = 0
                # push all args to the stack
                # keep track of how many bytes total we have pushed
                for arg_node in node_args:
                    dirs.extend(self.emit(arg_node, state))
                    arg_converted_type = state.contextual_types[arg_node]
                    arg_byte_count += arg_converted_type.getMaxSize()
                # then push cmd opcode to stack as u32
                dirs.append(
                    PushValDirective(FwOpcodeType(func.cmd.get_op_code()).serialize())
                )
                # now that all args are pushed to the stack, pop them and opcode off the stack
                # as a command
                dirs.append(StackCmdDirective(arg_byte_count))
        elif is_instance_compat(func, BuiltinSymbol):
            # put all arg values on stack
            for arg_node in node_args:
                dirs.extend(self.emit(arg_node, state))

            dirs.extend(func.generate(node))
        elif is_instance_compat(func, TypeCtorSymbol):
            # put arg values onto stack in correct order for serialization
            for arg_node in node_args:
                dirs.extend(self.emit(arg_node, state))
        elif is_instance_compat(func, CastSymbol):
            # just putting the arg value on the stack should be good enough, the
            # conversion will happen below
            dirs.extend(self.emit(node_args[0], state))
        elif is_instance_compat(func, FunctionSymbol):
            # script-defined function
            # okay.. calling convention says we're going to put the args on the stack
            for arg_node in node_args:
                dirs.extend(self.emit(arg_node, state))
            # okay, args are on the stack. now we're going to generate CALL
            func_entry_label = state.func_entry_labels[func.definition]
            # push the offset of the func
            dirs.append(IrPushLabelOffset(func_entry_label))
            # pop it off the stack and perform func call
            dirs.append(CallDirective())
        else:
            assert False, func

        # perform type conversion if called for
        unconverted_type = state.synthesized_types[node]
        converted_type = state.contextual_types[node]
        if unconverted_type != converted_type:
            dirs.extend(self.convert_numeric_type(unconverted_type, converted_type))

        return dirs

    def emit_AstAssign(self, node: AstAssign, state: CompileState):
        lhs = state.resolved_symbols[node.lhs]

        const_frame_offset = -1
        is_global_var = False
        if is_instance_compat(lhs, VariableSymbol):
            const_frame_offset = lhs.frame_offset
            is_global_var = lhs.is_global
            assert const_frame_offset is not None, lhs
        else:
            # okay now push the lvar arr offset to stack
            assert is_instance_compat(lhs, FieldSymbol), lhs
            assert is_instance_compat(lhs.base_sym, VariableSymbol), lhs.base_sym
            is_global_var = lhs.base_sym.is_global

            # is the lvar array offset a constant?
            # okay, are we assigning to a member or an element?
            if lhs.is_struct_member:
                # if it's a struct, then the lvar offset is always constant
                const_frame_offset = lhs.base_offset + lhs.base_sym.frame_offset
            else:
                assert lhs.is_array_element
                # again, offset is the offset in base type + offset of base lvar

                # however, because array idx can be variable, we might not know at compile time
                # the offset in base type.

                # check if we have a value for it
                const_idx_expr_value = state.contextual_values.get(lhs.idx_expr)
                if const_idx_expr_value is not None:
                    assert is_instance_compat(const_idx_expr_value, ArrayIndexType)
                    # okay, so we have a constant value index
                    lhs_parent_type = state.contextual_types[lhs.parent_expr]
                    const_frame_offset = (
                        lhs.base_sym.frame_offset
                        + const_idx_expr_value.val
                        * lhs_parent_type.MEMBER_TYPE.getMaxSize()
                    )
                # otherwise, the array idx is unknown at compile time. we will have to calculate it

        # Use global directives only when inside a function AND accessing a global variable
        use_global = self.in_function and is_global_var

        # start with rhs on stack
        dirs = self.emit(node.rhs, state)

        if const_frame_offset != -1:
            # in this case, we can use StoreConstOffset
            if use_global:
                dirs.append(
                    StoreAbsConstOffsetDirective(
                        const_frame_offset, lhs.type.getMaxSize()
                    )
                )
            else:
                dirs.append(
                    StoreRelConstOffsetDirective(
                        const_frame_offset, lhs.type.getMaxSize()
                    )
                )
        else:
            # okay we don't know the offset at compile time
            # only one case where that can be:
            assert is_instance_compat(lhs, FieldSymbol) and lhs.is_array_element, lhs

            # we need to calculate absolute offset in lvar array
            # == (parent offset) + (offset in parent)

            # offset in parent:
            lhs_parent_type = state.contextual_types[lhs.parent_expr]
            dirs.extend(
                self.calc_lvar_offset_of_array_element(
                    node, lhs.idx_expr, lhs_parent_type, state
                )
            )

            # parent offset:
            dirs.append(
                PushValDirective(U64Value(lhs.base_sym.frame_offset).serialize())
            )

            # add them
            dirs.append(IntAddDirective())

            # and now convert the u64 back into the I32Value that store expects
            dirs.extend(self.convert_numeric_type(U64Value, I32Value))

            # now that lvar array offset is pushed, use it to store in lvar array
            if use_global:
                dirs.append(StoreAbsDirective(lhs.type.getMaxSize()))
            else:
                dirs.append(StoreRelDirective(lhs.type.getMaxSize()))

        return dirs

    def emit_AstLiteral(self, node: AstLiteral, state: CompileState):
        const_dirs = self.try_emit_expr_as_const(node, state)
        assert const_dirs is not None
        return const_dirs

    def emit_AstAssert(self, node: AstAssert, state: CompileState):
        dirs = self.emit(node.condition, state)
        # invert the condition, we want to continue to exit if fail
        dirs.append(NotDirective())
        end_label = IrLabel(node, f"pass")
        dirs.append(IrIf(end_label))
        # push the error code we should use if false, if one was given
        if node.exit_code is not None:
            dirs.extend(self.emit(node.exit_code, state))
        else:
            # otherwise just use the default EXIT_WITH_ERROR error code
            dirs.append(
                PushValDirective(
                    U8Value(DirectiveErrorCode.EXIT_WITH_ERROR.value).serialize()
                )
            )
        dirs.append(ExitDirective())
        dirs.append(end_label)

        return dirs


class GenerateModule(Emitter):

    def emit_AstBlock(self, node: AstBlock, state: CompileState):
        if node is not state.root:
            return []

        # generate the main function using GenerateTopLevel (not in a function context)
        main_body = GenerateTopLevel().emit(node, state)

        # if there are functions, emit them at the top with a goto to skip past them
        if state.generated_funcs:
            funcs_code = []
            func_code_end_label = IrLabel(node, "main")
            funcs_code.append(IrGoto(func_code_end_label))
            for func, code in state.generated_funcs.items():
                funcs_code.extend(code)
            funcs_code.append(func_code_end_label)
            return funcs_code + main_body

        return main_body


class GenerateTopLevel(GenerateFunctionBody):
    """Generates top-level (main) code, not inside any function.
    At top level, stack_frame_start = 0, so local and global offsets are equivalent.
    All variables use LOCAL directives."""

    in_function = False


class IrPass:
    def run(
        self, ir: list[Directive | Ir], state: CompileState
    ) -> Union[list[Directive | Ir], BackendError]:
        pass


class ResolveLabels(IrPass):
    def run(self, ir, state: CompileState):
        labels: dict[str, int] = {}
        idx = 0
        dirs = []
        for dir in ir:
            if is_instance_compat(dir, IrLabel):
                if dir.name in labels:
                    return BackendError(f"Label {dir.name} already exists")
                labels[dir.name] = idx
                continue
            idx += 1

        # okay, we have all the labels
        for dir in ir:
            if is_instance_compat(dir, IrLabel):
                # drop these from the result
                continue
            elif is_instance_compat(dir, IrGoto):
                label = dir.label.name
                if label not in labels:
                    return BackendError(f"Unknown label {label}")
                dirs.append(GotoDirective(labels[label]))
            elif is_instance_compat(dir, IrIf):
                label = dir.goto_if_false_label.name
                if label not in labels:
                    return BackendError(f"Unknown label {label}")
                dirs.append(IfDirective(labels[label]))
            elif is_instance_compat(dir, IrPushLabelOffset):
                label = dir.label.name
                if label not in labels:
                    return BackendError(f"Unknown label {label}")
                dirs.append(PushValDirective(StackSizeType(labels[label]).serialize()))
            else:
                dirs.append(dir)

        return dirs


class FinalChecks(IrPass):
    def run(self, ir, state):
        if len(ir) > MAX_DIRECTIVES_COUNT:
            return BackendError(
                f"Too many directives in sequence (expected less than {MAX_DIRECTIVES_COUNT}, had {len(ir)})"
            )

        for dir in ir:
            # double check we've got rid of all the IR
            assert is_instance_compat(dir, Directive), dir

        return ir



from pathlib import Path
import tempfile
import traceback
import fpy.error
import fpy.model
from fpy.model import DirectiveErrorCode, FpySequencerModel
from fpy.bytecode.directives import AllocateDirective, Directive
from fpy.compiler import text_to_ast, ast_to_directives
from fprime_gds.common.loaders.ch_json_loader import ChJsonLoader
from fprime_gds.common.loaders.cmd_json_loader import CmdJsonLoader
from fprime_gds.common.loaders.event_json_loader import EventJsonLoader
from fprime_gds.common.loaders.prm_json_loader import PrmJsonLoader
from fprime_gds.common.testing_fw.api import IntegrationTestAPI


default_dictionary = str(
    Path(__file__).parent.parent.parent
    / "test"
    / "fpy"
    / "RefTopologyDictionary.json"
)


class CompilationFailed(Exception):
    """Raised when compilation fails expectedly (parse error or semantic error)."""
    pass


def compile_seq(fprime_test_api, seq: str, flags: list[str] = None) -> list[Directive]:
    """Compile a sequence string to a list of directives in memory."""
    fpy.error.file_name = "<test>"
    
    body = text_to_ast(seq)
    if body is None:
        # This shouldn't happen - text_to_ast calls exit(1) on parse errors
        raise CompilationFailed("Parsing failed")
    
    compile_args = {}
    for flag in flags or []:
        compile_args[flag] = True
    
    directives = ast_to_directives(body, default_dictionary, compile_args)
    if isinstance(directives, (fpy.error.CompileError, fpy.error.BackendError)):
        raise CompilationFailed(f"Compilation failed:\n{directives}")
    
    return directives


def lookup_type(fprime_test_api, type_name: str):
    dictionary = default_dictionary  # fprime_test_api.pipeline.dictionary_path
    cmd_json_dict_loader = CmdJsonLoader(dictionary)
    (cmd_id_dict, cmd_name_dict, versions) = cmd_json_dict_loader.construct_dicts(
        dictionary
    )

    ch_json_dict_loader = ChJsonLoader(dictionary)
    (ch_id_dict, ch_name_dict, versions) = ch_json_dict_loader.construct_dicts(
        dictionary
    )
    prm_json_dict_loader = PrmJsonLoader(dictionary)
    (prm_id_dict, prm_name_dict, versions) = prm_json_dict_loader.construct_dicts(
        dictionary
    )
    event_json_dict_loader = EventJsonLoader(dictionary)
    (event_id_dict, event_name_dict, versions) = event_json_dict_loader.construct_dicts(
        dictionary
    )
    type_name_dict = cmd_json_dict_loader.parsed_types
    type_name_dict.update(ch_json_dict_loader.parsed_types)
    type_name_dict.update(prm_json_dict_loader.parsed_types)
    type_name_dict.update(event_json_dict_loader.parsed_types)

    return type_name_dict[type_name]


def run_seq(
    fprime_test_api: IntegrationTestAPI,
    directives: list[Directive],
    tlm: dict[str, bytes] = None,
):
    """Run a list of directives using the sequencer model."""
    if tlm is None:
        tlm = {}

    # seq_file = tempfile.NamedTemporaryFile(suffix=".bin", delete=False)
    # Path(seq_file.name).write_bytes(serialize_directives(directives)[0])
    # fprime_test_api.send_and_assert_command("Ref.cmdSeq.RUN", [seq_file.name, "BLOCK"], timeout=4)
    # return

    dictionary = default_dictionary

    ch_json_dict_loader = ChJsonLoader(dictionary)
    (ch_id_dict, ch_name_dict, versions) = ch_json_dict_loader.construct_dicts(
        dictionary
    )
    cmd_json_dict_loader = CmdJsonLoader(dictionary)
    (cmd_id_dict, cmd_name_dict, versions) = cmd_json_dict_loader.construct_dicts(
        dictionary
    )
    fpy.model.debug = True
    model = FpySequencerModel(cmd_dict=cmd_id_dict)
    tlm_db = {}
    for chan_name, val in tlm.items():
        ch_template = ch_name_dict[chan_name]
        tlm_db[ch_template.get_id()] = val
    ret = model.run(directives, tlm_db)
    if ret != DirectiveErrorCode.NO_ERROR:
        raise RuntimeError("Sequence returned", ret)
    if len(directives) > 0 and isinstance(directives[0], AllocateDirective):
        # check that the start and end sizes are the same
        if len(model.stack) != directives[0].size:
            raise RuntimeError(f"Sequence leaked {len(model.stack) - directives[0].size} bytes")


def assert_compile_success(fprime_test_api, seq: str, flags: list[str] = None):
    compile_seq(fprime_test_api, seq, flags)


def assert_run_success(fprime_test_api, seq: str, tlm: dict[str, bytes] = None, flags: list[str] = None):
    directives = compile_seq(fprime_test_api, seq, flags)
    run_seq(fprime_test_api, directives, tlm)


def assert_compile_failure(fprime_test_api, seq: str, flags: list[str] = None):
    try:
        compile_seq(fprime_test_api, seq, flags)
    except (SystemExit, CompilationFailed):
        # Compilation failed as expected
        return

    # no error was generated
    raise RuntimeError("compile_seq succeeded")


def assert_run_failure(fprime_test_api, seq: str, flags: list[str] = None):
    directives = compile_seq(fprime_test_api, seq, flags)
    try:
        run_seq(fprime_test_api, directives)
    except (RuntimeError, AssertionError) as e:
        print(e)
        return

    # other exceptions we will let through, such as assertions
    raise RuntimeError("run_seq succeeded")



from pathlib import Path

import pytest

from fpy import main as fpy_main
import fpy.error as fpy_error
import fpy.model as fpy_model


@pytest.mark.parametrize(
    "size,expected",
    [
        (0, "0 B"),
        (512, "512 B"),
        (1024, "1 KB"),
        (1536, "1 KB"),
        (5 * 1024 * 1024, "5 MB"),
    ],
)
def test_human_readable_size(size, expected):
    assert fpy_main.human_readable_size(size) == expected


def test_compile_main_missing_input(tmp_path, capsys):
    missing = tmp_path / "missing.fpy"
    dict_path = tmp_path / "dict.json"
    with pytest.raises(SystemExit) as exc:
        fpy_main.compile_main(
            [
                str(missing),
                "--dictionary",
                str(dict_path),
            ]
        )
    assert exc.value.code == 1
    captured = capsys.readouterr()
    assert "does not exist" in captured.out


def test_compile_main_bytecode_output(monkeypatch, tmp_path, capsys):
    input_path = tmp_path / "seq.fpy"
    input_path.write_text("content")
    dict_path = tmp_path / "dict.json"
    dict_path.write_text("{}")

    monkeypatch.setattr(fpy_error, "debug", False, raising=False)
    monkeypatch.setattr(fpy_main, "text_to_ast", lambda text: "AST")

    def fake_ast_to_directives(body, dictionary):
        assert body == "AST"
        assert Path(dictionary) == dict_path
        return ["directive"]

    monkeypatch.setattr(fpy_main, "ast_to_directives", fake_ast_to_directives)
    monkeypatch.setattr(fpy_main, "directives_to_fpybc", lambda directives: "FPYBC")

    def fail_serialize(_):
        raise AssertionError("serialize_directives should not be called")

    monkeypatch.setattr(fpy_main, "serialize_directives", fail_serialize)

    fpy_main.compile_main(
        [
            str(input_path),
            "--dictionary",
            str(dict_path),
            "--bytecode",
            "--debug",
        ]
    )

    captured = capsys.readouterr()
    assert captured.out.strip() == "FPYBC"
    assert fpy_error.debug is True


def test_compile_main_binary_output(monkeypatch, tmp_path, capsys):
    input_path = tmp_path / "seq.fpy"
    input_path.write_text("content")
    dict_path = tmp_path / "dict.json"
    dict_path.write_text("{}")

    monkeypatch.setattr(fpy_main, "text_to_ast", lambda text: "AST")
    monkeypatch.setattr(
        fpy_main,
        "ast_to_directives",
        lambda body, dictionary: ["directive"],
    )
    monkeypatch.setattr(fpy_main, "directives_to_fpybc", lambda directives: "FPYBC")
    monkeypatch.setattr(
        fpy_main,
        "serialize_directives",
        lambda directives: (b"\x01\x02", 0xABCD),
    )

    fpy_main.compile_main(
        [
            str(input_path),
            "--dictionary",
            str(dict_path),
        ]
    )

    output_path = input_path.with_suffix(".bin")
    assert output_path.read_bytes() == b"\x01\x02"
    captured = capsys.readouterr()
    assert "CRC 0xabcd" in captured.out
    assert "2 B" in captured.out


def test_model_main_success(monkeypatch, tmp_path):
    binary = tmp_path / "seq.bin"
    binary.write_bytes(b"data")

    monkeypatch.setattr(fpy_model, "debug", False, raising=False)
    monkeypatch.setattr(fpy_main, "deserialize_directives", lambda data: ["dir"])

    instances = []

    class DummyModel:
        def __init__(self):
            instances.append(self)
            self.ran_with = None

        def run(self, directives):
            self.ran_with = directives
            return fpy_main.DirectiveErrorCode.NO_ERROR

    monkeypatch.setattr(fpy_main, "FpySequencerModel", DummyModel)

    fpy_main.model_main([str(binary), "--debug"])

    assert fpy_model.debug is True
    assert instances[0].ran_with == ["dir"]


def test_model_main_failure(monkeypatch, tmp_path, capsys):
    binary = tmp_path / "seq.bin"
    binary.write_bytes(b"data")

    monkeypatch.setattr(fpy_main, "deserialize_directives", lambda data: ["dir"])

    class DummyModel:
        def run(self, directives):
            return fpy_main.DirectiveErrorCode.EXIT_WITH_ERROR

    monkeypatch.setattr(fpy_main, "FpySequencerModel", DummyModel)

    with pytest.raises(SystemExit) as exc:
        fpy_main.model_main([str(binary)])

    assert exc.value.code == 1
    captured = capsys.readouterr()
    assert "Sequence failed" in captured.out


def test_assemble_main_missing_input(tmp_path, capsys):
    source = tmp_path / "seq.fpybc"
    with pytest.raises(SystemExit) as exc:
        fpy_main.assemble_main([str(source)])
    assert exc.value.code == 1
    captured = capsys.readouterr()
    assert "does not exist" in captured.out


def test_assemble_main_writes_binary(monkeypatch, tmp_path, capsys):
    source = tmp_path / "seq.fpybc"
    source.write_text("bc")

    monkeypatch.setattr(fpy_main, "fpybc_parse", lambda text: ["body"])
    monkeypatch.setattr(fpy_main, "assemble", lambda body: ["dirs"])
    monkeypatch.setattr(
        fpy_main,
        "serialize_directives",
        lambda directives: (b"\x03\x04\x05", 0x1234),
    )

    fpy_main.assemble_main([str(source)])

    output_path = source.with_suffix(".bin")
    assert output_path.read_bytes() == b"\x03\x04\x05"
    captured = capsys.readouterr()
    assert "CRC 0x1234" in captured.out


def test_disassemble_main_missing_input(tmp_path, capsys):
    source = tmp_path / "seq.bin"
    with pytest.raises(SystemExit) as exc:
        fpy_main.disassemble_main([str(source)])
    assert exc.value.code == 1
    captured = capsys.readouterr()
    assert "does not exist" in captured.out


def test_disassemble_main_writes_text(monkeypatch, tmp_path, capsys):
    source = tmp_path / "seq.bin"
    source.write_bytes(b"data")

    monkeypatch.setattr(fpy_main, "deserialize_directives", lambda data: ["dirs"])
    monkeypatch.setattr(fpy_main, "directives_to_fpybc", lambda dirs: "FPYBC")

    fpy_main.disassemble_main([str(source)])

    output_path = source.with_suffix(".fpybc")
    assert output_path.read_text() == "FPYBC"
    captured = capsys.readouterr()
    assert captured.out.strip() == "Done"



from __future__ import annotations
import sys
from functools import lru_cache
from pathlib import Path
from fprime_gds.common.models.serialize.time_type import TimeType as TimeValue
from fprime_gds.common.models.serialize.bool_type import BoolType as BoolValue
from fprime_gds.common.models.serialize.enum_type import EnumType as EnumValue
from fprime_gds.common.models.serialize.serializable_type import (
    SerializableType as StructValue,
)
from fprime_gds.common.models.serialize.array_type import ArrayType as ArrayValue
from fprime_gds.common.models.serialize.numerical_types import (
    U8Type as U8Value,
    U16Type as U16Value,
    U32Type as U32Value,
    NumericalType as NumericalValue,
)
from fprime_gds.common.models.serialize.type_base import BaseType as FppValue
from lark import Lark
from fpy.bytecode.directives import Directive
from fpy.codegen import (
    AssignVariableOffsets,
    FinalChecks,
    GenerateFunctionEntryPoints,
    GenerateFunctions,
    GenerateModule,
    IrPass,
    ResolveLabels,
)
from fpy.desugaring import DesugarDefaultArgs, DesugarForLoops
from fpy.semantics import (
    AssignIds,
    AssignLocalScopes,
    CalculateConstExprValues,
    CalculateDefaultArgConstValues,
    CheckBreakAndContinueInLoop,
    CheckConstArrayAccesses,
    CheckFunctionReturns,
    CheckReturnInFunc,
    CheckUseBeforeDeclare,
    CreateVariablesAndFuncs,
    PickTypesAndResolveAttrsAndItems,
    ResolveTypeNames,
    ResolveVars,
    WarnRangesAreNotEmpty,
)
from fpy.syntax import AstBlock, FpyTransformer, PythonIndenter
from fpy.macros import MACROS
from fpy.types import (
    SPECIFIC_NUMERIC_TYPES,
    CompileState,
    FppType,
    CallableSymbol,
    CastSymbol,
    CommandSymbol,
    TypeCtorSymbol,
    Visitor,
    create_symbol_table,
)
from fprime_gds.common.loaders.ch_json_loader import ChJsonLoader
from fprime_gds.common.loaders.cmd_json_loader import CmdJsonLoader
from fprime_gds.common.loaders.event_json_loader import EventJsonLoader
from fprime_gds.common.loaders.prm_json_loader import PrmJsonLoader
from fprime_gds.common.templates.cmd_template import CmdTemplate
from pathlib import Path
from lark import Lark, LarkError

from fpy.error import BackendError, CompileError, handle_lark_error
import fpy.error

# Load grammar once at module level
_fpy_grammar_path = Path(__file__).parent / "grammar.lark"
_fpy_grammar_str = _fpy_grammar_path.read_text()

# Create parser once at module level with LALR and cache enabled.
# PythonIndenter.process() resets its internal state on each call,
# so it's safe to reuse the same parser instance.
_fpy_indenter = PythonIndenter()
_fpy_parser = Lark(
    _fpy_grammar_str,
    start="input",
    parser="lalr",
    postlex=_fpy_indenter,
    propagate_positions=True,
    maybe_placeholders=True,
)


def text_to_ast(text: str):
    from lark.exceptions import VisitError

    fpy.error.input_text = text
    fpy.error.input_lines = text.splitlines()
    try:
        tree = _fpy_parser.parse(text, on_error=handle_lark_error)
    except LarkError as e:
        handle_lark_error(e)
        return None
    try:
        transformed = FpyTransformer().transform(tree)
    except RecursionError:
        print(fpy.error.CompileError("Maximum recursion depth exceeded (code is too deeply nested)"), file=sys.stderr)
        exit(1)
    except VisitError as e:
        # VisitError wraps exceptions that occur during tree transformation
        if isinstance(e.orig_exc, RecursionError):
            print(fpy.error.CompileError("Maximum recursion depth exceeded (code is too deeply nested)"), file=sys.stderr)
        else:
            print(fpy.error.CompileError(f"Internal error during parsing: {e.orig_exc}"), file=sys.stderr)
        exit(1)
    return transformed


@lru_cache(maxsize=4)
def _load_dictionary(dictionary: str) -> tuple:
    """
    Load and parse the dictionary file once, caching the results.
    Returns a tuple of (cmd_name_dict, ch_name_dict, prm_name_dict, type_name_dict).
    """
    cmd_json_dict_loader = CmdJsonLoader(dictionary)
    (_, cmd_name_dict, _) = cmd_json_dict_loader.construct_dicts(dictionary)

    ch_json_dict_loader = ChJsonLoader(dictionary)
    (_, ch_name_dict, _) = ch_json_dict_loader.construct_dicts(dictionary)
    prm_json_dict_loader = PrmJsonLoader(dictionary)
    (_, prm_name_dict, _) = prm_json_dict_loader.construct_dicts(dictionary)
    event_json_dict_loader = EventJsonLoader(dictionary)
    (_, _, _) = event_json_dict_loader.construct_dicts(dictionary)

    # the type name dict is a mapping of a fully qualified name to an fprime type
    # here we put into it all types found while parsing all cmds, params and tlm channels
    type_name_dict: dict[str, FppType] = dict(cmd_json_dict_loader.parsed_types)
    type_name_dict.update(ch_json_dict_loader.parsed_types)
    type_name_dict.update(prm_json_dict_loader.parsed_types)
    type_name_dict.update(event_json_dict_loader.parsed_types)

    return (cmd_name_dict, ch_name_dict, prm_name_dict, type_name_dict)


@lru_cache(maxsize=4)
def _build_scopes(dictionary: str) -> tuple:
    """
    Build and cache the scopes for a dictionary.
    Returns tuple of (tlm_scope, prm_scope, type_scope, callable_scope, const_scope).
    """
    cmd_name_dict, ch_name_dict, prm_name_dict, type_name_dict = _load_dictionary(dictionary)
    
    # Make a copy of type_name_dict since we'll mutate it
    type_name_dict = dict(type_name_dict)

    # enum const dict is a dict of fully qualified enum const name (like Ref.Choice.ONE) to its fprime value
    enum_const_name_dict: dict[str, FppValue] = {}

    # find each enum type, and put each of its values in the enum const dict
    for name, typ in type_name_dict.items():
        if issubclass(typ, EnumValue):
            for enum_const_name, val in typ.ENUM_DICT.items():
                enum_const_name_dict[name + "." + enum_const_name] = typ(
                    enum_const_name
                )

    # insert the builtin types into the dict
    type_name_dict["Fw.Time"] = TimeValue
    for typ in SPECIFIC_NUMERIC_TYPES:
        type_name_dict[typ.get_canonical_name()] = typ
    type_name_dict["bool"] = BoolValue
    # note no string type at the moment

    cmd_response_type = type_name_dict["Fw.CmdResponse"]
    callable_name_dict: dict[str, CallableSymbol] = {}
    # add all cmds to the callable dict
    for name, cmd in cmd_name_dict.items():
        cmd: CmdTemplate
        args = []
        for arg_name, _, arg_type in cmd.arguments:
            args.append((arg_name, arg_type, None))  # No default values for cmds
        # cmds are thought of as callables with a Fw.CmdResponse return value
        callable_name_dict[name] = CommandSymbol(
            cmd.get_full_name(), cmd_response_type, args, cmd
        )

    # add numeric type casts to callable dict
    for typ in SPECIFIC_NUMERIC_TYPES:
        callable_name_dict[typ.get_canonical_name()] = CastSymbol(
            typ.get_canonical_name(), typ, [("value", NumericalValue, None)], typ
        )

    # for each type in the dict, if it has a constructor, create an TypeCtorSymbol
    # object to track the constructor and put it in the callable name dict
    for name, typ in type_name_dict.items():
        args = []
        if issubclass(typ, StructValue):
            for arg_name, arg_type, _, _ in typ.MEMBER_LIST:
                args.append(
                    (arg_name, arg_type, None)
                )  # No default values for struct ctors
        elif issubclass(typ, ArrayValue):
            for i in range(0, typ.LENGTH):
                args.append(("e" + str(i), typ.MEMBER_TYPE, None))
        elif issubclass(typ, TimeValue):
            args.append(("time_base", U16Value, None))
            args.append(("time_context", U8Value, None))
            args.append(("seconds", U32Value, None))
            args.append(("useconds", U32Value, None))
        else:
            # bool, enum, string or numeric type
            # none of these have callable ctors
            continue

        callable_name_dict[name] = TypeCtorSymbol(name, typ, args, typ)

    # for each macro function, add it to the callable dict
    for macro_name, macro in MACROS.items():
        callable_name_dict[macro_name] = macro

    return (
        create_symbol_table(ch_name_dict),
        create_symbol_table(prm_name_dict),
        create_symbol_table(type_name_dict),
        create_symbol_table(callable_name_dict),
        create_symbol_table(enum_const_name_dict),
    )


def get_base_compile_state(dictionary: str, compile_args: dict) -> CompileState:
    """return the initial state of the compiler, based on the given dict path"""
    tlm_scope, prm_scope, type_scope, callable_scope, const_scope = _build_scopes(dictionary)

    state = CompileState(
        tlms=tlm_scope,
        prms=prm_scope,
        types=type_scope,
        callables=callable_scope,
        consts=const_scope,
        compile_args=compile_args or dict(),
    )
    return state


def ast_to_directives(
    body: AstBlock,
    dictionary: str,
    compile_args: dict | None = None,
) -> list[Directive] | CompileError | BackendError:
    compile_args = compile_args or dict()
    state = get_base_compile_state(dictionary, compile_args)
    state.root = body
    semantics_passes: list[Visitor] = [
        # assign each node a unique id for indexing/hashing
        AssignIds(),
        # based on position of node in tree, figure out which scope it is in
        AssignLocalScopes(),
        # based on assignment syntax nodes, we know which variables exist where
        CreateVariablesAndFuncs(),
        # check that break/continue are in loops, and store which loop they're in
        CheckBreakAndContinueInLoop(),
        CheckReturnInFunc(),
        # resolve type annotations first, since they use a restricted syntax (AstTypeExpr)
        # and we need to know variable types before resolving other references
        ResolveTypeNames(),
        # resolve all variable and function references
        ResolveVars(),
        # make sure we don't use any variables before they are declared
        CheckUseBeforeDeclare(),
        # this pass resolves all attributes and items, as well as determines the type of expressions
        PickTypesAndResolveAttrsAndItems(),
        # Calculate const values for default arguments first (and check they're const).
        # This must happen before CalculateConstExprValues because call sites may
        # reference functions defined later in the source, and we need the default
        # values' const values to be available.
        CalculateDefaultArgConstValues(),
        # okay, now that we're sure we're passing in all the right args to each func,
        # we can calculate values of type ctors etc etc
        CalculateConstExprValues(),
        CheckFunctionReturns(),
        CheckConstArrayAccesses(),
        WarnRangesAreNotEmpty(),
    ]
    desugaring_passes: list[Visitor] = [
        # Fill in default arguments before desugaring for loops
        DesugarDefaultArgs(),
        # now that semantic analysis is done, we can desugar things. start with for loops
        DesugarForLoops(),
    ]
    codegen_passes = [
        # Assign variable offsets before generating function bodies
        # so global variable offsets are known when referenced in functions
        AssignVariableOffsets(),
        GenerateFunctionEntryPoints(),
        # generate all function bodies
        GenerateFunctions(),
    ]
    module_generator = GenerateModule()

    ir_passes: list[IrPass] = [ResolveLabels(), FinalChecks()]

    for compile_pass in semantics_passes:
        compile_pass.run(body, state)
        if len(state.errors) != 0:
            return state.errors[0]

    for compile_pass in desugaring_passes:
        compile_pass.run(body, state)
        if len(state.errors) != 0:
            return state.errors[0]

    for compile_pass in codegen_passes:
        compile_pass.run(body, state)
        if len(state.errors) != 0:
            return state.errors[0]

    ir = module_generator.emit(body, state)

    for compile_pass in ir_passes:
        ir = compile_pass.run(ir, state)
        if isinstance(ir, BackendError):
            # early return errors
            return ir

    # print out warnings
    for warning in state.warnings:
        print(warning)

    # all the ir is guaranteed to have been converted to directives by now
    return ir



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Iterator, List, Literal as TypingLiteral, Union
from lark import Token, Transformer, v_args
from lark.tree import Meta
from lark.lark import PostLex
from lark.indenter import DedentError
from decimal import Decimal


class PythonIndenter(PostLex):
    # from lark, but slightly modified to fix a bug
    """This is a postlexer that "injects" indent/dedent tokens based on indentation.

    It keeps track of the current indentation, as well as the current level of parentheses.
    Inside parentheses, the indentation is ignored, and no indent/dedent tokens get generated.
    See also: the ``postlex`` option in `Lark`.
    """
    paren_level: int
    indent_level: List[int]
    NL_type = "_NEWLINE"
    OPEN_PAREN_types = ["LPAR", "LSQB", "LBRACE"]
    CLOSE_PAREN_types = ["RPAR", "RSQB", "RBRACE"]
    INDENT_type = "_INDENT"
    DEDENT_type = "_DEDENT"
    tab_len = 8

    def __init__(self) -> None:
        self.paren_level = 0
        self.indent_level = [0]
        assert self.tab_len > 0

    def handle_NL(self, token: Token) -> Iterator[Token]:
        if self.paren_level > 0:
            return

        yield token

        if not "\n" in token:
            return

        indent_str = token.rsplit("\n", 1)[1]  # Tabs and spaces
        indent = indent_str.count(" ") + indent_str.count("\t") * self.tab_len

        if indent > self.indent_level[-1]:
            self.indent_level.append(indent)
            yield Token.new_borrow_pos(self.INDENT_type, indent_str, token)
        else:
            while indent < self.indent_level[-1]:
                self.indent_level.pop()
                yield Token.new_borrow_pos(self.DEDENT_type, indent_str, token)

            if indent != self.indent_level[-1]:
                raise DedentError(
                    "Unexpected dedent to column %s. Expected dedent to %s"
                    % (indent, self.indent_level[-1])
                )

    def _process(self, stream):
        for token in stream:
            if token.type == self.NL_type:
                yield from self.handle_NL(token)
            else:
                yield token

            if token.type in self.OPEN_PAREN_types:
                self.paren_level += 1
            elif token.type in self.CLOSE_PAREN_types:
                self.paren_level -= 1
                assert self.paren_level >= 0

        while len(self.indent_level) > 1:
            self.indent_level.pop()
            yield Token(self.DEDENT_type, "")

        assert self.indent_level == [0], self.indent_level

    def process(self, stream):
        self.paren_level = 0
        self.indent_level = [0]
        return self._process(stream)


@dataclass
class Ast:
    meta: Meta = field(repr=False)
    id: int = field(init=False, repr=False, default=None)

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, value):
        if not isinstance(value, Ast):
            return False
        assert self.id is not None
        return self.id == value.id


@dataclass
class AstVar(Ast):
    var: str


@dataclass
class AstTypeExpr(Ast):
    """A qualified name like A.b.c - used for type annotations"""
    parts: list[str]


@dataclass()
class AstString(Ast):
    value: str


@dataclass
class AstNumber(Ast):
    value: int | Decimal


@dataclass
class AstBoolean(Ast):
    value: TypingLiteral[True] | TypingLiteral[False]


AstLiteral = Union[AstString, AstNumber, AstBoolean]


@dataclass
class AstMemberAccess(Ast):
    parent: "AstExpr"
    attr: str


@dataclass
class AstIndexExpr(Ast):
    parent: "AstExpr"
    item: AstExpr


@dataclass
class AstNamedArgument(Ast):
    name: str
    value: "AstExpr"


@dataclass
class AstFuncCall(Ast):
    func: "AstExpr"
    # args can contain both positional (AstExpr) and named arguments (AstNamedArgument)
    args: list[Union["AstExpr", AstNamedArgument]] | None


@dataclass
class AstPass(Ast):
    pass  # ha ha


@dataclass
class AstBinaryOp(Ast):
    lhs: AstExpr
    op: str
    rhs: AstExpr


@dataclass
class AstUnaryOp(Ast):
    op: str
    val: AstExpr


@dataclass
class AstRange(Ast):
    lower_bound: AstExpr
    op: str
    upper_bound: AstExpr


AstOp = Union[AstBinaryOp, AstUnaryOp]

AstReference = Union[AstMemberAccess, AstIndexExpr, AstVar]
AstExpr = Union[AstFuncCall, AstLiteral, AstReference, AstOp, AstRange]


@dataclass
class AstAssign(Ast):
    lhs: AstExpr
    type_ann: AstTypeExpr | None
    rhs: AstExpr


@dataclass
class AstElif(Ast):
    condition: AstExpr
    body: "AstStmtList"


@dataclass()
class AstIf(Ast):
    condition: AstExpr
    body: "AstStmtList"
    elifs: list[AstElif]
    els: Union["AstStmtList", None]


@dataclass
class AstFor(Ast):
    loop_var: AstVar
    range: AstExpr
    body: AstStmtList


@dataclass
class AstWhile(Ast):
    condition: AstExpr
    body: AstStmtList


@dataclass
class AstAssert(Ast):
    condition: AstExpr
    exit_code: Union[AstExpr, None]


@dataclass
class AstBreak(Ast):
    pass


@dataclass
class AstContinue(Ast):
    pass


@dataclass
class AstReturn(Ast):
    value: Union[AstExpr, None]


@dataclass
class AstDef(Ast):
    name: AstVar
    # parameters is a list of (name, type, default_value) tuples
    # default_value is None if no default is provided
    parameters: list[tuple[AstVar, AstTypeExpr, AstExpr | None]]
    return_type: Union[AstTypeExpr, None]
    body: AstBlock


AstStmt = Union[
    AstExpr,
    AstAssign,
    AstPass,
    AstIf,
    AstElif,
    AstFor,
    AstBreak,
    AstContinue,
    AstWhile,
    AstAssert,
    AstDef,
    AstReturn
]
AstStmtWithExpr = Union[
    AstExpr, AstAssign, AstIf, AstElif, AstFor, AstWhile, AstAssert, AstDef, AstReturn
]
AstNodeWithSideEffects = Union[
    AstFuncCall,
    AstAssign,
    AstIf,
    AstElif,
    AstFor,
    AstWhile,
    AstAssert,
    AstBreak,
    AstContinue,
    AstDef,
    AstReturn
]


@dataclass
class AstStmtList(Ast):
    stmts: list[AstStmt]


@dataclass
class AstBlock(Ast):
    stmts: list[AstStmt]


for cls in Ast.__subclasses__():
    cls.__hash__ = Ast.__hash__
    # cls.__repr__ = Ast.__repr__


@v_args(meta=False, inline=False)
def as_list(self, tree):
    return list(tree)


def no_inline_or_meta(type):
    @v_args(meta=False, inline=False)
    def wrapper(self, tree):
        return type(tree)

    return wrapper


def no_inline(type):
    @v_args(meta=True, inline=False)
    def wrapper(self, meta, tree):
        return type(meta, tree)

    return wrapper


def no_meta(type):
    @v_args(meta=False, inline=True)
    def wrapper(self, tree):
        return type(tree)

    return wrapper


def handle_str(meta, s: str):
    return s.strip("'").strip('"')


def handle_parameter(meta, args):
    """Parse a single parameter: (name, type, default_value or None)"""
    assert len(args) in (2, 3), f"Expected 2 or 3 args, got {len(args)}: {args}"
    name, type_expr = args[0], args[1]
    default_value = args[2] if len(args) == 3 else None
    return (name, type_expr, default_value)


@v_args(meta=True, inline=True)
class FpyTransformer(Transformer):
    input = no_inline(AstBlock)
    pass_stmt = AstPass

    assign = AstAssign

    for_stmt = AstFor
    while_stmt = AstWhile
    block = no_inline(AstBlock)
    break_stmt = AstBreak
    continue_stmt = AstContinue

    assert_stmt = AstAssert

    if_stmt = AstIf
    elifs = no_inline_or_meta(list)
    elif_ = AstElif
    stmt_list = no_inline(AstStmtList)
    binary_op = AstBinaryOp
    unary_op = AstUnaryOp

    func_call = AstFuncCall
    arguments = no_inline_or_meta(list)
    named_argument = AstNamedArgument

    @v_args(meta=True, inline=True)
    def positional_argument(self, meta, value):
        # Just return the expression directly for positional arguments
        return value

    string = AstString
    number = AstNumber
    boolean = AstBoolean
    name = no_meta(str)
    member_access = AstMemberAccess
    index_expr = AstIndexExpr
    var = AstVar
    range = AstRange

    type_expr = no_inline(AstTypeExpr)

    def_stmt = AstDef
    parameter = no_inline(handle_parameter)
    parameters = no_inline_or_meta(list)  # Just convert to list
    return_stmt = AstReturn

    NAME = str
    DEC_NUMBER = int
    FLOAT_NUMBER = Decimal
    COMPARISON_OP = str
    RANGE_OP = str
    STRING = handle_str
    CONST_TRUE = lambda a, b: True
    CONST_FALSE = lambda a, b: False
    ADD_OP: str
    SUB_OP: str
    DIV_OP: str
    MUL_OP: str
    FLOOR_DIV_OP: str
    MOD_OP: str
    POW_OP: str



from __future__ import annotations
from abc import ABC
from decimal import Decimal
import inspect
from dataclasses import astuple, dataclass, field, fields
import math
import struct
import typing
from typing import Callable, Iterable, Union, get_args, get_origin
import itertools
import zlib

from fpy.error import CompileError
from fpy.ir import Ir, IrLabel

# In Python 3.10+, the `|` operator creates a `types.UnionType`.
# We need to handle this for forward compatibility, but it won't exist in 3.9.
try:
    from types import UnionType

    UNION_TYPES = (Union, UnionType)
except ImportError:
    UNION_TYPES = (Union,)

from fpy.bytecode.directives import (
    Directive,
)
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime_gds.common.templates.prm_template import PrmTemplate
from fprime_gds.common.models.serialize.numerical_types import (
    U8Type as U8Value,
    U16Type as U16Value,
    U32Type as U32Value,
    U64Type as U64Value,
    I8Type as I8Value,
    I16Type as I16Value,
    I32Type as I32Value,
    I64Type as I64Value,
    F32Type as F32Value,
    F64Type as F64Value,
    IntegerType as IntegerValue,
    FloatType as FloatValue,
    NumericalType as NumericalValue,
)
from fprime_gds.common.models.serialize.string_type import StringType as StringValue
from fpy.syntax import (
    AstBreak,
    AstContinue,
    AstDef,
    AstExpr,
    AstFor,
    AstFuncCall,
    AstTypeExpr,
    AstOp,
    AstReference,
    Ast,
    AstAssign,
    AstReturn,
    AstBlock,
    AstWhile,
)
from fprime_gds.common.models.serialize.type_base import BaseType as FppValue

MAX_DIRECTIVES_COUNT = 1024
MAX_DIRECTIVE_SIZE = 2048

COMPILER_MAX_STRING_SIZE = 128


def typename(typ: FppType) -> str:
    if typ == FpyIntegerValue:
        return "Integer"
    if typ == FpyFloatValue:
        return "Float"
    if issubclass(typ, NumericalValue):
        return typ.get_canonical_name()
    if typ == FpyStringValue:
        return "String"
    if typ == RangeValue:
        return "Range"
    return str(typ)


# this is the "internal" integer type that integer literals have by
# default. it is arbitrary precision. it is also only used in places where
# we know the value is constant
class FpyIntegerValue(IntegerValue):
    @classmethod
    def range(cls):
        raise NotImplementedError()

    @staticmethod
    def get_serialize_format():
        raise NotImplementedError()

    @classmethod
    def get_bits(cls):
        return math.inf

    @classmethod
    def validate(cls, val):
        if not isinstance(val, int):
            raise RuntimeError()


# this is the "internal" float type that float literals have by
# default. it is arbitrary precision. it is also only used in places where
# we know the value is constant
class FpyFloatValue(FloatValue):
    @staticmethod
    def get_serialize_format():
        raise NotImplementedError()

    @classmethod
    def get_bits(cls):
        return math.inf

    @classmethod
    def validate(cls, val):
        if not isinstance(val, Decimal):
            raise RuntimeError()


class RangeValue(FppValue):
    """the type produced by range expressions `X .. Y`"""
    def serialize(self):
        raise NotImplementedError()

    def deserialize(self, data, offset):
        raise NotImplementedError()

    def getSize(self):
        raise NotImplementedError()

    @classmethod
    def getMaxSize(cls):
        raise NotImplementedError()

    def __repr__(self):
        return self.__class__.__name__

    def to_jsonable(self):
        raise NotImplementedError()


# this is the "internal" string type that string literals have by
# default. it is arbitrary length. it is also only used in places where
# we know the value is constant
FpyStringValue = StringValue.construct_type("FpyStringValue", None)

SPECIFIC_NUMERIC_TYPES = (
    U32Value,
    U16Value,
    U64Value,
    U8Value,
    I16Value,
    I32Value,
    I64Value,
    I8Value,
    F32Value,
    F64Value,
)
SPECIFIC_INTEGER_TYPES = (
    U32Value,
    U16Value,
    U64Value,
    U8Value,
    I16Value,
    I32Value,
    I64Value,
    I8Value,
)
SIGNED_INTEGER_TYPES = (
    I16Value,
    I32Value,
    I64Value,
    I8Value,
)
UNSIGNED_INTEGER_TYPES = (
    U32Value,
    U16Value,
    U64Value,
    U8Value,
)
SPECIFIC_FLOAT_TYPES = (
    F32Value,
    F64Value,
)
ARBITRARY_PRECISION_TYPES = (FpyFloatValue, FpyIntegerValue)


def is_instance_compat(obj, cls):
    """
    A wrapper for isinstance() that correctly handles Union types in Python 3.9+.

    Args:
        obj: The object to check.
        cls: The class, tuple of classes, or Union type to check against.

    Returns:
        True if the object is an instance of the class or any type in the Union.
    """
    origin = get_origin(cls)
    if origin in UNION_TYPES:
        # It's a Union type, so get its arguments.
        # e.g., get_args(Union[int, str]) returns (int, str)
        return isinstance(obj, get_args(cls))

    # It's not a Union, so it's a regular type (like int) or a
    # tuple of types ((int, str)), which isinstance handles natively.
    return isinstance(obj, cls)


# a value of type FppType is a Python `type` object representing
# the type of an Fprime value
FppType = type[FppValue]


class NothingValue(ABC):
    """a type which has no valid values in fprime. used to denote
    a function which doesn't return a value"""

    @classmethod
    def __subclasscheck__(cls, subclass):
        return False


# the `type` object representing the NothingType class
NothingType = type[NothingValue]


@dataclass
class CallableSymbol:
    name: str
    return_type: FppType | NothingType
    # args is a list of (name, type, default_value) tuples
    # default_value is an AstExpr or None if no default is provided
    args: list[tuple[str, FppType, "AstExpr | None"]]


@dataclass
class CommandSymbol(CallableSymbol):
    cmd: CmdTemplate


@dataclass
class BuiltinSymbol(CallableSymbol):
    generate: Callable[[AstFuncCall], list[Directive]]
    """a function which instantiates the builtin given the calling node"""

@dataclass
class FunctionSymbol(CallableSymbol):
    definition: AstDef


@dataclass
class TypeCtorSymbol(CallableSymbol):
    type: FppType


@dataclass
class CastSymbol(CallableSymbol):
    to_type: FppType


@dataclass
class FieldSymbol:
    """a reference to a member/element of an fprime struct/array type"""

    parent_expr: AstExpr
    """the complete qualifier"""
    base_sym: Union["Symbol", None]
    """the base symbol, up through all the layers of field symbols, or None if parent at some point is not a symbol at all"""
    type: FppType
    """the fprime type of this reference"""
    is_struct_member: bool = False
    """True if this is a struct member reference"""
    is_array_element: bool = False
    """True if this is an array element reference"""
    base_offset: int = None
    """the constant offset in the base symbol type, or None if unknown at compile time"""
    local_offset: int = None
    """the constant offset in the parent type at which to find this field
    or None if unknown at compile time"""
    name: str = None
    """the name of the field, if applicable"""
    idx_expr: AstExpr = None
    """the expression that evaluates to the index in the parent array of the field, if applicable"""


# named variables can be tlm chans, prms, callables, or directly referenced consts (usually enums)
@dataclass
class VariableSymbol:
    """a mutable, typed value stored on the stack referenced by an unqualified name"""

    name: str
    type_ref: AstTypeExpr | None
    """the expression denoting the var's type"""
    declaration: Ast
    """the node where this var is declared"""
    type: FppType | None = None
    """the resolved type of the variable. None if type unsure at the moment"""
    frame_offset: int | None = None
    """the offset in the lvar array where this var is stored"""
    is_global: bool = False
    """whether this variable is a top-level (global) variable"""


@dataclass
class ForLoopAnalysis:
    loop_var: VariableSymbol
    upper_bound_var: VariableSymbol
    reuse_existing_loop_var: bool
    

next_symbol_table_id = 0


# a symbol table (scope) 
class SymbolTable(dict):
    def __init__(self):
        global next_symbol_table_id
        self.id = next_symbol_table_id
        next_symbol_table_id += 1

    def __getitem__(self, key: str) -> Symbol:
        return super().__getitem__(key)

    def get(self, key) -> Symbol | None:
        return super().get(key, None)

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, value):
        return isinstance(value, SymbolTable) and value.id == self.id


def create_symbol_table(
    symbols: dict[str, "Symbol"],
) -> SymbolTable:
    """from a flat dict of strs to symbols, creates a hierarchical symbol table.
    no two leaf nodes may have the same name"""

    base = SymbolTable()

    for fqn, sym in symbols.items():
        names_strs = fqn.split(".")

        ns = base
        while len(names_strs) > 1:
            existing_child = ns.get(names_strs[0])
            if existing_child is None:
                # this symbol table is not defined atm
                existing_child = {}
                ns[names_strs[0]] = existing_child

            if not isinstance(existing_child, dict):
                # something else already has this name
                break

            ns = existing_child
            names_strs = names_strs[1:]

        if len(names_strs) != 1:
            # broke early. skip this loop
            continue

        # okay, now ns is the complete scope of the attribute
        # i.e. everything up until the last '.'
        name = names_strs[0]

        existing_child = ns.get(name)

        if existing_child is not None:
            # uh oh, something already had this name with a diff value
            continue

        ns[name] = sym

    return base


def merge_symbol_tables(lhs: SymbolTable, rhs: SymbolTable) -> SymbolTable:
    """returns the two symbol tables, joined into one. if there is a conflict, chooses lhs over rhs"""
    lhs_keys = set(lhs.keys())
    rhs_keys = set(rhs.keys())
    common_keys = lhs_keys.intersection(rhs_keys)

    only_lhs_keys = lhs_keys.difference(common_keys)
    only_rhs_keys = rhs_keys.difference(common_keys)

    new = SymbolTable()

    for key in common_keys:
        if not isinstance(lhs[key], dict) or not isinstance(rhs[key], dict):
            # cannot be merged cleanly. one of the two is not a symbol table
            new[key] = lhs[key]
            continue

        new[key] = merge_symbol_tables(lhs[key], rhs[key])

    for key in only_lhs_keys:
        new[key] = lhs[key]
    for key in only_rhs_keys:
        new[key] = rhs[key]

    return new


Symbol = typing.Union[
    ChTemplate,
    PrmTemplate,
    FppValue,
    CallableSymbol,
    FppType,
    VariableSymbol,
    FieldSymbol,
    SymbolTable
]
"""a named entity in fpy that can be looked up in a symbol table"""


def lookup_symbol(node: Ast, name: str, state: CompileState) -> VariableSymbol:
    """look up a symbol by name, searching this scope and all parent scopes"""
    symbol_table = state.local_scopes[node]
    resolved = None
    while symbol_table is not None and resolved is None:
        resolved = symbol_table.get(name)
        symbol_table = state.scope_parents[symbol_table]

    return resolved


@dataclass
class CompileState:
    """a collection of input, internal and output state variables and maps"""

    types: SymbolTable
    """a symbol table whose leaf nodes are subclasses of BaseType"""
    callables: SymbolTable
    """a symbol table whose leaf nodes are CallableSymbol instances"""
    tlms: SymbolTable
    """a symbol table whose leaf nodes are ChTemplates"""
    prms: SymbolTable
    """a symbol table whose leaf nodes are PrmTemplates"""
    consts: SymbolTable
    """a symbol table whose leaf nodes are VariableSymbols"""
    runtime_values: SymbolTable = None
    """a symbol table whose leaf nodes are tlms/prms/consts, all of which
    have some value at runtime."""

    compile_args: dict = field(default_factory=dict)

    def __post_init__(self):
        self.runtime_values = merge_symbol_tables(
            self.tlms,
            merge_symbol_tables(self.prms, self.consts),
        )

    next_node_id: int = 0
    root: AstBlock = None
    scope_parents: dict[AstBlock, AstBlock | None] = field(
        default_factory=dict, repr=False
    )
    """map of a scoped body node to the parent scoped body node it should use"""
    local_scopes: dict[Ast, SymbolTable] = field(default_factory=dict, repr=False)
    """map of node to the SymbolTable it should resolve names in"""
    for_loops: dict[AstFor, ForLoopAnalysis] = field(default_factory=dict)
    """map of for loops to a ForLoopAnalysis struct, which contains additional info about the loops"""
    enclosing_loops: dict[Union[AstBreak, AstContinue], Union[AstFor, AstWhile]] = (
        field(default_factory=dict)
    )
    """map of break/continue to the loop which contains the break/continue"""
    desugared_for_loops: dict[AstWhile, AstFor] = field(default_factory=dict)
    """mapping of while loops which are desugared for loops, to the original node from which they came"""

    enclosing_funcs: dict[AstReturn, AstDef] = field(default_factory=dict)

    resolved_symbols: dict[AstReference, Symbol] = field(
        default_factory=dict, repr=False
    )
    """reference to its singular resolution"""

    synthesized_types: dict[AstExpr, FppType | NothingType] = field(
        default_factory=dict
    )
    """expr to its fprime type, before type conversions are applied"""

    op_intermediate_types: dict[AstOp, FppType] = field(default_factory=dict)
    """the intermediate type that all args should be converted to for the given op"""

    expr_explicit_casts: list[AstExpr] = field(default_factory=list)
    """a list of nodes which are explicit casts"""
    contextual_types: dict[AstExpr, FppType] = field(default_factory=dict)
    """expr to fprime type it will end up being on the stack after type conversions"""

    contextual_values: dict[AstExpr, FppValue | NothingValue | None] = field(
        default_factory=dict
    )
    """expr to the fprime value it will end up being on the stack after type conversions.
    None if unsure at compile time"""

    resolved_func_args: dict[AstFuncCall, list[AstExpr]] = field(
        default_factory=dict
    )
    """function call to resolved arguments in positional order.
    Default values are filled in for arguments not provided at the call site."""

    while_loop_end_labels: dict[AstWhile, IrLabel] = field(default_factory=dict)
    """while loop node mapped to the label pointing to the end of the loop"""
    while_loop_start_labels: dict[AstWhile, IrLabel] = field(default_factory=dict)
    """while loop node mapped to the label pointing to the start of the loop, just before the conditional"""
    # store keys as while because for loops are desugared to while
    for_loop_inc_labels: dict[AstWhile, IrLabel] = field(default_factory=dict)
    """for loop node (desugared into a while) mapped to a label pointing to its increment stmt"""

    does_return: dict[Ast, bool] = field(default_factory=dict)

    func_entry_labels: dict[AstDef, IrLabel] = field(default_factory=dict)
    """function to entry point label"""

    generated_funcs: dict[AstDef, list[Directive|Ir]] = field(default_factory=dict)

    errors: list[CompileError] = field(default_factory=list)
    """a list of all compile exceptions generated by passes"""

    warnings: list[CompileError] = field(default_factory=list)
    """a list of all compiler warnings generated by passes"""

    next_anon_var_id: int = 0

    def new_anonymous_variable_name(self) -> str:
        id = self.next_anon_var_id
        self.next_anon_var_id += 1
        return f"$value{id}"

    def err(self, msg, n):
        """adds a compile exception to internal state"""
        self.errors.append(CompileError(msg, n))

    def warn(self, msg, n):
        self.warnings.append(CompileError("Warning: " + msg, n))


# Cache for visitor method mappings, keyed by visitor class
_visitor_cache: dict[type, dict[type, str]] = {}


class Visitor:
    """visits each class, calling a custom visit function, if one is defined, for each
    node type"""

    def __init__(self):
        self.visitors: dict[type[Ast], Callable] = {}
        """dict of node type to handler function"""
        self.build_visitor_dict()

    def build_visitor_dict(self):
        cls = type(self)
        # Check if this class's visitor mapping is already cached
        if cls in _visitor_cache:
            # Use cached mapping (maps node type -> method name)
            for node_type, method_name in _visitor_cache[cls].items():
                self.visitors[node_type] = getattr(self, method_name)
            return

        # Build the mapping and cache it
        class_cache: dict[type, str] = {}
        for name, func in inspect.getmembers(cls, inspect.isfunction):
            if not name.startswith("visit") or name == "visit_default":
                # not a visitor, or the default visit func
                continue
            signature = inspect.signature(func)
            params = list(signature.parameters.values())
            assert len(params) == 3
            assert params[1].annotation is not None
            annotations = typing.get_type_hints(func)
            param_type = annotations[params[1].name]

            origin = get_origin(param_type)
            if origin in UNION_TYPES:
                # It's a Union type, so get its arguments.
                for t in get_args(param_type):
                    class_cache[t] = name
                    self.visitors[t] = getattr(self, name)
            else:
                # It's not a Union, so it's a regular type
                class_cache[param_type] = name
                self.visitors[param_type] = getattr(self, name)

        _visitor_cache[cls] = class_cache

    def _visit(self, node: Ast, state: CompileState):
        visit_func = self.visitors.get(type(node), self.visit_default)
        return visit_func(node, state)

    def visit_default(self, node: Ast, state: CompileState):
        pass

    def run(self, start: Ast, state: CompileState):
        """runs the visitor, starting at the given node, descending depth-first"""

        def _descend(node: Ast):
            if not isinstance(node, Ast):
                return
            children = []
            for field in fields(node):
                field_val = getattr(node, field.name)
                if isinstance(field_val, list):
                    # also handle the one case where we have a list of tuples
                    if len(field_val) > 0 and isinstance(field_val[0], tuple):
                        field_val = itertools.chain.from_iterable(field_val)
                    children.extend(field_val)
                else:
                    children.append(field_val)

            for child in children:
                if not isinstance(child, Ast):
                    continue
                _descend(child)
                if len(state.errors) != 0:
                    break
                self._visit(child, state)
                if len(state.errors) != 0:
                    break

        _descend(start)
        self._visit(start, state)


class TopDownVisitor(Visitor):

    def run(self, start: Ast, state: CompileState):
        """runs the visitor, starting at the given node, descending breadth-first"""

        def _descend(node: Ast):
            if not isinstance(node, Ast):
                return
            children = []
            for field in fields(node):
                field_val = getattr(node, field.name)
                if isinstance(field_val, list):
                    # also handle the one case where we have a list of tuples
                    if len(field_val) > 0 and isinstance(field_val[0], tuple):
                        field_val = itertools.chain.from_iterable(field_val)
                    children.extend(field_val)
                else:
                    children.append(field_val)

            for child in children:
                if not isinstance(child, Ast):
                    continue
                self._visit(child, state)
                if len(state.errors) != 0:
                    break
                _descend(child)
                if len(state.errors) != 0:
                    break

        self._visit(start, state)
        _descend(start)


class Transformer(Visitor):

    class Delete:
        pass

    def run(self, start: Ast, state: CompileState):

        def _descend(node):
            if not isinstance(node, Ast):
                return
            for field in fields(node):
                field_val = getattr(node, field.name)
                if isinstance(field_val, list):
                    # child is a list, iterate over each member of the list
                    # use a copy so we can remove as we traverse, also so
                    # we don't visit things that we added

                    #
                    idx = -1
                    for child in field_val[:]:
                        idx += 1
                        if not isinstance(child, Ast):
                            continue
                        _descend(child)
                        if len(state.errors) != 0:
                            break
                        transformed = self._visit(child, state)
                        if len(state.errors) != 0:
                            break
                        if isinstance(transformed, Iterable):
                            assert all(
                                isinstance(n, Ast) for n in transformed
                            ), transformed
                            # func split one node into many
                            # remove the original child and add the new ones
                            # insert them in the place where the child used to be, in the right order
                            field_val.remove(child)
                            for new_child_idx, new_child in enumerate(transformed):
                                field_val.insert(idx + new_child_idx, new_child)
                            # make sure that we maintain insertion order by updating the idx
                            # accounting for our removal of an original node
                            # if we don't do this, then if we were to insert into list after this based on idx,
                            # the positions could be swapped around
                            idx += len(transformed) - 1
                        elif isinstance(transformed, Ast):
                            field_val.remove(child)
                            field_val.insert(idx, transformed)
                        elif transformed is Transformer.Delete:
                            # just delete it
                            field_val.remove(child)
                        else:
                            assert transformed is None, transformed
                            # don't do anything, didn't return anything
                    if len(state.errors) != 0:
                        # need a second check here to get out of the enclosing loop
                        break
                    # don't need to update the field, it was a ptr to a list so should
                    # already be updated
                else:
                    _descend(field_val)
                    if len(state.errors) != 0:
                        break
                    transformed = self._visit(field_val, state)
                    if len(state.errors) != 0:
                        break
                    if isinstance(transformed, Ast):
                        setattr(node, field.name, transformed)
                    elif transformed is Transformer.Delete:
                        # just delete it
                        setattr(node, field.name, None)
                    else:
                        # cannot return a list if the original attr wasn't a list
                        assert transformed is None, transformed
                        # don't do anything, didn't return anything

        _descend(start)
        self._visit(start, state)


# Cache for emitter method mappings, keyed by emitter class
_emitter_cache: dict[type, dict[type, str]] = {}


class Emitter:
    # Default: not in a function (top-level code)
    # Subclasses override this to indicate function body context
    in_function = False

    def __init__(self):
        self.emitters: dict[type[Ast], Callable] = {}
        """dict of node type to handler function"""
        self.build_emitter_dict()

    def build_emitter_dict(self):
        cls = type(self)
        # Check if this class's emitter mapping is already cached
        if cls in _emitter_cache:
            # Use cached mapping (maps node type -> method name)
            for node_type, method_name in _emitter_cache[cls].items():
                self.emitters[node_type] = getattr(self, method_name)
            return

        # Build the mapping and cache it
        class_cache: dict[type, str] = {}
        for name, func in inspect.getmembers(cls, inspect.isfunction):
            if not name.startswith("emit_"):
                # not an emitter
                continue
            signature = inspect.signature(func)
            params = list(signature.parameters.values())
            assert len(params) == 3
            assert params[1].annotation is not None
            annotations = typing.get_type_hints(func)
            param_type = annotations[params[1].name]

            origin = get_origin(param_type)
            if origin in UNION_TYPES:
                # It's a Union type, so get its arguments.
                for t in get_args(param_type):
                    class_cache[t] = name
                    self.emitters[t] = getattr(self, name)
            else:
                # It's not a Union, so it's a regular type
                class_cache[param_type] = name
                self.emitters[param_type] = getattr(self, name)

        _emitter_cache[cls] = class_cache

    def emit(self, node: Ast, state: CompileState) -> list[Directive | Ir]:
        return self.emitters[type(node)](node, state)


MAJOR_VERSION = 0
MINOR_VERSION = 3
PATCH_VERSION = 0
SCHEMA_VERSION = 4

HEADER_FORMAT = "!BBBBBHI"
HEADER_SIZE = struct.calcsize(HEADER_FORMAT)


@dataclass
class Header:
    majorVersion: int
    minorVersion: int
    patchVersion: int
    schemaVersion: int
    argumentCount: int
    statementCount: int
    bodySize: int


FOOTER_FORMAT = "!I"
FOOTER_SIZE = struct.calcsize(FOOTER_FORMAT)


@dataclass
class Footer:
    crc: int


def deserialize_directives(bytes: bytes) -> list[Directive]:
    header = Header(*struct.unpack_from(HEADER_FORMAT, bytes))

    if header.schemaVersion != SCHEMA_VERSION:
        raise RuntimeError(
            f"Schema version wrong (expected {SCHEMA_VERSION} found {header.schemaVersion})"
        )

    dirs = []
    idx = 0
    offset = HEADER_SIZE
    while idx < header.statementCount:
        offset_and_dir = Directive.deserialize(bytes, offset)
        if offset_and_dir is None:
            raise RuntimeError("Unable to deserialize sequence")
        offset, dir = offset_and_dir
        dirs.append(dir)
        idx += 1

    if offset != len(bytes) - FOOTER_SIZE:
        raise RuntimeError(
            f"{len(bytes) - FOOTER_SIZE - offset} extra bytes at end of sequence"
        )

    return dirs


def serialize_directives(dirs: list[Directive]) -> tuple[bytes, int]:
    output_bytes = bytes()

    for dir in dirs:
        dir_bytes = dir.serialize()
        if len(dir_bytes) > MAX_DIRECTIVE_SIZE:
            print(
                CompileError(
                    f"Directive {dir} in sequence too large (expected less than {MAX_DIRECTIVE_SIZE}, was {len(dir_bytes)})"
                )
            )
            exit(1)
        output_bytes += dir_bytes

    header = Header(
        MAJOR_VERSION,
        MINOR_VERSION,
        PATCH_VERSION,
        SCHEMA_VERSION,
        0,
        len(dirs),
        len(output_bytes),
    )
    output_bytes = struct.pack(HEADER_FORMAT, *astuple(header)) + output_bytes

    crc = zlib.crc32(output_bytes) % (1 << 32)
    footer = Footer(crc)
    output_bytes += struct.pack(FOOTER_FORMAT, *astuple(footer))

    return output_bytes, crc



from __future__ import annotations
import typing
from typing import Any

# This makes the code forward-compatible. In Python 3.10+, the `|` operator
# creates a types.UnionType. In 3.9, only typing.Union exists.
try:
    from types import UnionType

    UNION_TYPES = (typing.Union, UnionType)
except ImportError:
    UNION_TYPES = (typing.Union,)

from dataclasses import dataclass, fields
from typing import ClassVar
import typing
from typing import Union
import struct
from fprime_gds.common.models.serialize.type_base import BaseType as FppValue
from fprime_gds.common.models.serialize.numerical_types import (
    U8Type as U8Value,
    U16Type as U16Value,
    U32Type as U32Value,
    U64Type as U64Value,
    I8Type as I8Value,
    I16Type as I16Value,
    I32Type as I32Value,
    I64Type as I64Value,
    F32Type as F32Value,
    F64Type as F64Value,
)
from fprime_gds.common.models.serialize.bool_type import BoolType as BoolValue
from enum import Enum

FwSizeType = U64Value
FwChanIdType = U32Value
FwPrmIdType = U32Value
FwOpcodeType = U32Value
ArrayIndexType = I64Value
StackSizeType = U32Value
LoopVarType = ArrayIndexType


def get_union_members(type_hint: type) -> list[type]:
    """
    If the type_hint is a Union, returns a list of its member types.
    Otherwise, returns the original type_hint.
    """
    # get_origin returns the base type (e.g., Union for Union[int, str])
    # or None if it's a simple type like int.
    origin = typing.get_origin(type_hint)

    if origin in UNION_TYPES:
        # get_args returns the type arguments (e.g., (int, str))
        return list(typing.get_args(type_hint))

    # Not a Union, so return the type itself
    return [type_hint]


class DirectiveId(Enum):
    INVALID = 0
    WAIT_REL = 1
    WAIT_ABS = 2
    GOTO = 3
    IF = 4
    NO_OP = 5
    PUSH_TLM_VAL = 6
    PUSH_PRM = 7
    CONST_CMD = 8
    # stack op directives
    # all of these are handled at the CPP level by one StackOpDirective to save boilerplate
    # you MUST keep them all in between OR and ITRUNC_64_32 inclusive
    # boolean ops
    OR = 9
    AND = 10
    # integer equalities
    IEQ = 11
    INE = 12
    # unsigned integer inequalities
    ULT = 13
    ULE = 14
    UGT = 15
    UGE = 16
    # signed integer inequalities
    SLT = 17
    SLE = 18
    SGT = 19
    SGE = 20
    # floating point equalities
    FEQ = 21
    FNE = 22
    # floating point inequalities
    FLT = 23
    FLE = 24
    FGT = 25
    FGE = 26
    NOT = 27
    # floating point conversion to signed/unsigned integer,
    # and vice versa
    FPTOSI = 28
    FPTOUI = 29
    SITOFP = 30
    UITOFP = 31
    # integer arithmetic
    ADD = 32
    SUB = 33
    MUL = 34
    UDIV = 35
    SDIV = 36
    UMOD = 37
    SMOD = 38
    # float arithmetic
    FADD = 39
    FSUB = 40
    FMUL = 41
    FDIV = 42
    FPOW = 43
    FLOG = 44
    FMOD = 45
    # floating point bitwidth conversions
    FPEXT = 46
    FPTRUNC = 47
    # integer bitwidth conversions
    # signed integer extend
    SIEXT_8_64 = 48
    SIEXT_16_64 = 49
    SIEXT_32_64 = 50
    # zero (unsigned) integer extend
    ZIEXT_8_64 = 51
    ZIEXT_16_64 = 52
    ZIEXT_32_64 = 53
    # integer truncate
    ITRUNC_64_8 = 54
    ITRUNC_64_16 = 55
    ITRUNC_64_32 = 56
    # end stack op dirs

    EXIT = 57
    ALLOCATE = 58
    STORE_REL_CONST_OFFSET = 59
    LOAD_REL = 60
    PUSH_VAL = 61
    DISCARD = 62
    MEMCMP = 63
    STACK_CMD = 64
    PUSH_TLM_VAL_AND_TIME = 65
    PUSH_TIME = 66
    SET_FLAG = 67
    GET_FLAG = 68
    GET_FIELD = 69
    PEEK = 70
    STORE_REL = 71
    CALL = 72
    RETURN = 73
    LOAD_ABS = 74
    STORE_ABS = 75
    STORE_ABS_CONST_OFFSET = 76


class Directive:
    opcode: ClassVar[DirectiveId] = DirectiveId.INVALID

    def serialize(self) -> bytes:
        arg_bytes = self.serialize_args()

        output = U8Value(self.opcode.value).serialize()
        output += U16Value(len(arg_bytes)).serialize()
        output += arg_bytes

        return output

    def serialize_args(self) -> bytes:
        output = bytes()

        for field in fields(self):
            value = getattr(self, field.name)
            if isinstance(value, FppValue):
                # it is already an fprime type instance
                # so we can serialize it
                output += value.serialize()
                continue

            if isinstance(value, bytes):
                # it is just raw bytes
                output += value
                continue

            # okay, it is not a primitive type or bytes
            field_type = typing.get_type_hints(self.__class__)[field.name]
            union_members = get_union_members(field_type)
            primitive_type = None
            # find out which primitive type it is
            for arg in union_members:
                if issubclass(arg, FppValue):
                    # it is a primitive type
                    primitive_type = arg
                    break
            if primitive_type is None:
                raise NotImplementedError(
                    "Unknown how to serialize field", field.name, "for", self
                )

            output += primitive_type(value).serialize()

        return output

    def __repr__(self):
        r = self.__class__.__old_repr__(self)
        name = self.__class__.__name__.replace("Directive", "").upper()
        value = "".join(r.split("(")[1:])
        return name + "(" + value

    @classmethod
    def deserialize(cls, data: bytes, offset: int) -> tuple[int, "Directive"] | None:
        if len(data) - offset < 3:
            # insufficient space
            return None
        opcode = struct.unpack_from(">B", data, offset)[0]
        arg_size = struct.unpack_from(">H", data, offset + 1)[0]
        offset += 3
        if len(data) - offset < arg_size:
            # insufficient space
            return None
        args = data[offset : (offset + arg_size)]
        offset += arg_size
        dir_type = [
            c
            for c in (Directive.__subclasses__() + StackOpDirective.__subclasses__())
            if c.opcode.value == opcode
        ]
        if len(dir_type) != 1:
            return None

        arg_offset = 0
        dir_type = dir_type[0]
        arg_values = []

        # go through each field in the type of the directive
        for field in fields(dir_type):
            field_type = typing.get_type_hints(dir_type)[field.name]
            # get a list of all union members of the field type
            # or a list containing just the type if it is not a union
            union_types = get_union_members(field_type)

            base_type = None
            for t in union_types:
                if issubclass(t, FppValue):
                    base_type = t

            # if one of the members of the union was a sub of basetype
            if base_type is not None:
                # deserialize using that basetype and add to arg value list
                instance = base_type()
                instance.deserialize(args, arg_offset)
                arg_values.append(instance.val)
                arg_offset += instance.getSize()
                continue
            # none of the args were base types. the only other thing we could be
            # is a byte array. assert that that's true
            assert len(union_types) == 1 and union_types[0] == bytes
            # it is just raw bytes. deserialize until the end
            arg_values.append(args[arg_offset:])
            arg_offset = len(args)
            continue

        dir = dir_type(*arg_values)
        return offset, dir


@dataclass
class StackOpDirective(Directive):
    """the argument types this dir pops off the stack"""

    stack_output_type: ClassVar[type[FppValue]] = FppValue
    """the type this dir pushes to the stack"""


@dataclass
class StackCmdDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.STACK_CMD

    args_size: Union[int, StackSizeType]


@dataclass
class MemCompareDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.MEMCMP
    size: Union[int, StackSizeType]


@dataclass
class LoadRelDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.LOAD_REL

    lvar_offset: Union[int, I32Value]
    size: Union[int, StackSizeType]


@dataclass
class IntegerSignedExtend8To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SIEXT_8_64
    stack_output_type: ClassVar[type[FppValue]] = I64Value


@dataclass
class IntegerSignedExtend16To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SIEXT_16_64
    stack_output_type: ClassVar[type[FppValue]] = I64Value


@dataclass
class IntegerSignedExtend32To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SIEXT_32_64
    stack_output_type: ClassVar[type[FppValue]] = I64Value


@dataclass
class IntegerZeroExtend8To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ZIEXT_8_64
    stack_output_type: ClassVar[type[FppValue]] = U64Value


@dataclass
class IntegerZeroExtend16To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ZIEXT_16_64
    stack_output_type: ClassVar[type[FppValue]] = U64Value


@dataclass
class IntegerZeroExtend32To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ZIEXT_32_64
    stack_output_type: ClassVar[type[FppValue]] = U64Value


@dataclass
class IntegerTruncate64To8Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ITRUNC_64_8
    stack_output_type: ClassVar[type[FppValue]] = I8Value


@dataclass
class IntegerTruncate64To16Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ITRUNC_64_16
    stack_output_type: ClassVar[type[FppValue]] = I16Value


@dataclass
class IntegerTruncate64To32Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ITRUNC_64_32
    stack_output_type: ClassVar[type[FppValue]] = I32Value


@dataclass
class AllocateDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.ALLOCATE

    size: Union[int, StackSizeType]


@dataclass
class StoreRelDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.STORE_REL

    size: Union[int, StackSizeType]


@dataclass
class StoreRelConstOffsetDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.STORE_REL_CONST_OFFSET

    lvar_offset: Union[int, I32Value]
    size: Union[int, StackSizeType]


@dataclass
class DiscardDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.DISCARD

    size: Union[int, StackSizeType]


@dataclass
class PushValDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.PUSH_VAL

    val: bytes


@dataclass
class ConstCmdDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.CONST_CMD

    cmd_opcode: Union[int, FwOpcodeType]
    args: bytes


@dataclass
class FloatModuloDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FMOD
    stack_output_type: ClassVar[type[FppValue]] = F64Value


@dataclass
class SignedModuloDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SMOD
    stack_output_type: ClassVar[type[FppValue]] = I64Value


@dataclass
class UnsignedModuloDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UMOD
    stack_output_type: ClassVar[type[FppValue]] = U64Value


@dataclass
class IntAddDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ADD
    stack_output_type: ClassVar[type[FppValue]] = I64Value


@dataclass
class IntSubtractDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SUB
    stack_output_type: ClassVar[type[FppValue]] = I64Value


@dataclass
class IntMultiplyDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.MUL
    stack_output_type: ClassVar[type[FppValue]] = I64Value


@dataclass
class UnsignedIntDivideDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UDIV
    stack_output_type: ClassVar[type[FppValue]] = U64Value


@dataclass
class SignedIntDivideDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SDIV
    stack_output_type: ClassVar[type[FppValue]] = I64Value


@dataclass
class FloatAddDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FADD
    stack_output_type: ClassVar[type[FppValue]] = F64Value


@dataclass
class FloatSubtractDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FSUB
    stack_output_type: ClassVar[type[FppValue]] = F64Value


@dataclass
class FloatMultiplyDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FMUL
    stack_output_type: ClassVar[type[FppValue]] = F64Value


@dataclass
class FloatExponentDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPOW
    stack_output_type: ClassVar[type[FppValue]] = F64Value


@dataclass
class FloatDivideDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FDIV
    stack_output_type: ClassVar[type[FppValue]] = F64Value


@dataclass
class FloatLogDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FLOG
    stack_output_type: ClassVar[type[FppValue]] = F64Value


@dataclass
class WaitRelDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.WAIT_REL
    # seconds and useconds are implicit


@dataclass
class WaitAbsDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.WAIT_ABS
    # time base, time context, seconds and useconds are implicit


@dataclass
class GotoDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.GOTO
    dir_idx: Union[int, U32Value]


@dataclass
class IfDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.IF
    false_goto_dir_index: Union[int, U32Value]
    """U32: The dir index to go to if the top of stack is false."""


@dataclass
class NoOpDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.NO_OP


@dataclass
class PushTlmValDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.PUSH_TLM_VAL
    chan_id: Union[int, FwChanIdType]
    """FwChanIdType: The telemetry channel ID to get."""


@dataclass
class PushPrmDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.PUSH_PRM
    prm_id: Union[int, FwPrmIdType]
    """FwPrmIdType: The parameter ID to get the value of."""


@dataclass
class OrDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.OR
    stack_args: ClassVar[list[type[FppValue]]] = [BoolValue, BoolValue]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class AndDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.AND
    stack_args: ClassVar[list[type[FppValue]]] = [BoolValue, BoolValue]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class IntEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.IEQ
    stack_args: ClassVar[list[type[FppValue]]] = [
        Union[I64Value, U64Value],
        Union[I64Value, U64Value],
    ]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class IntNotEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.INE
    stack_args: ClassVar[list[type[FppValue]]] = [
        Union[I64Value, U64Value],
        Union[I64Value, U64Value],
    ]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class UnsignedLessThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ULT
    stack_args: ClassVar[list[type[FppValue]]] = [U64Value, U64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class UnsignedLessThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ULE
    stack_args: ClassVar[list[type[FppValue]]] = [U64Value, U64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class UnsignedGreaterThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UGT
    stack_args: ClassVar[list[type[FppValue]]] = [U64Value, U64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class UnsignedGreaterThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UGE
    stack_args: ClassVar[list[type[FppValue]]] = [U64Value, U64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class SignedLessThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SLT
    stack_args: ClassVar[list[type[FppValue]]] = [I64Value, I64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class SignedLessThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SLE
    stack_args: ClassVar[list[type[FppValue]]] = [I64Value, I64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class SignedGreaterThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SGT
    stack_args: ClassVar[list[type[FppValue]]] = [I64Value, I64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class SignedGreaterThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SGE
    stack_args: ClassVar[list[type[FppValue]]] = [I64Value, I64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class FloatGreaterThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FGE
    stack_args: ClassVar[list[type[FppValue]]] = [F64Value, F64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class FloatLessThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FLE
    stack_args: ClassVar[list[type[FppValue]]] = [F64Value, F64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class FloatLessThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FLT
    stack_args: ClassVar[list[type[FppValue]]] = [F64Value, F64Value]
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class FloatGreaterThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FGT
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class FloatEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FEQ
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class FloatNotEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FNE
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class NotDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.NOT
    stack_output_type: ClassVar[type[FppValue]] = BoolValue


@dataclass
class FloatTruncateDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPTRUNC
    stack_output_type: ClassVar[type[FppValue]] = F32Value


@dataclass
class FloatExtendDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPEXT
    stack_output_type: ClassVar[type[FppValue]] = F64Value


@dataclass
class FloatToSignedIntDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPTOSI
    stack_output_type: ClassVar[type[FppValue]] = I64Value


@dataclass
class SignedIntToFloatDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SITOFP
    stack_output_type: ClassVar[type[FppValue]] = F64Value


@dataclass
class FloatToUnsignedIntDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPTOUI
    stack_output_type: ClassVar[type[FppValue]] = U64Value


@dataclass
class UnsignedIntToFloatDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UITOFP
    stack_output_type: ClassVar[type[FppValue]] = F64Value
    # src implied


@dataclass
class ExitDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.EXIT


@dataclass
class GetFieldDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.GET_FIELD
    # pops an offset off the stack
    parent_size: Union[int, StackSizeType]
    member_size: Union[int, StackSizeType]


@dataclass
class PeekDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.PEEK


@dataclass
class PushTimeDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.PUSH_TIME


@dataclass
class CallDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.CALL


@dataclass
class ReturnDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.RETURN

    return_val_size: Union[int, StackSizeType]
    call_args_size: Union[int, StackSizeType]


@dataclass
class LoadAbsDirective(Directive):
    """Load a value from a global variable (absolute offset from start of stack)"""
    opcode: ClassVar[DirectiveId] = DirectiveId.LOAD_ABS

    global_offset: Union[int, I32Value]
    size: Union[int, StackSizeType]


@dataclass
class StoreAbsDirective(Directive):
    """Store a value to a global variable (absolute offset popped from stack)"""
    opcode: ClassVar[DirectiveId] = DirectiveId.STORE_ABS

    size: Union[int, StackSizeType]


@dataclass
class StoreAbsConstOffsetDirective(Directive):
    """Store a value to a global variable at a constant absolute offset"""
    opcode: ClassVar[DirectiveId] = DirectiveId.STORE_ABS_CONST_OFFSET

    global_offset: Union[int, I32Value]
    size: Union[int, StackSizeType]


for cls in Directive.__subclasses__():
    cls.__old_repr__ = cls.__repr__
    cls.__repr__ = Directive.__repr__

for cls in StackOpDirective.__subclasses__():
    cls.__old_repr__ = cls.__repr__
    cls.__repr__ = StackOpDirective.__repr__


class UnaryStackOp(str, Enum):
    NOT = "not"
    IDENTITY = "+"
    NEGATE = "-"


class BinaryStackOp(str, Enum):
    EXPONENT = "**"
    MODULUS = "%"
    ADD = "+"
    SUBTRACT = "-"
    MULTIPLY = "*"
    DIVIDE = "/"
    FLOOR_DIVIDE = "//"
    GREATER_THAN = ">"
    GREATER_THAN_OR_EQUAL = ">="
    LESS_THAN_OR_EQUAL = "<="
    LESS_THAN = "<"
    EQUAL = "=="
    NOT_EQUAL = "!="
    OR = "or"
    AND = "and"


NUMERIC_OPERATORS = {
    UnaryStackOp.IDENTITY,
    UnaryStackOp.NEGATE,
    BinaryStackOp.ADD,
    BinaryStackOp.SUBTRACT,
    BinaryStackOp.MULTIPLY,
    BinaryStackOp.DIVIDE,
    BinaryStackOp.MODULUS,
    BinaryStackOp.EXPONENT,
    BinaryStackOp.FLOOR_DIVIDE,
}
BOOLEAN_OPERATORS = {UnaryStackOp.NOT, BinaryStackOp.OR, BinaryStackOp.AND}

UNARY_STACK_OPS: dict[str, dict[type[FppValue], type[StackOpDirective]]] = {
    UnaryStackOp.NOT: {BoolValue: NotDirective},
    UnaryStackOp.IDENTITY: {
        I64Value: NoOpDirective,
        U64Value: NoOpDirective,
        F64Value: NoOpDirective,
    },
    UnaryStackOp.NEGATE: {
        I64Value: IntMultiplyDirective,
        U64Value: IntMultiplyDirective,
        F64Value: FloatMultiplyDirective,
    },
}

BINARY_STACK_OPS: dict[str, dict[type[FppValue], type[StackOpDirective]]] = {
    BinaryStackOp.EXPONENT: {F64Value: FloatExponentDirective},
    BinaryStackOp.MODULUS: {
        I64Value: SignedModuloDirective,
        U64Value: UnsignedModuloDirective,
        F64Value: FloatModuloDirective,
    },
    BinaryStackOp.ADD: {
        I64Value: IntAddDirective,
        U64Value: IntAddDirective,
        F64Value: FloatAddDirective,
    },
    BinaryStackOp.SUBTRACT: {
        I64Value: IntSubtractDirective,
        U64Value: IntSubtractDirective,
        F64Value: FloatSubtractDirective,
    },
    BinaryStackOp.MULTIPLY: {
        I64Value: IntMultiplyDirective,
        U64Value: IntMultiplyDirective,
        F64Value: FloatMultiplyDirective,
    },
    BinaryStackOp.DIVIDE: {
        I64Value: SignedIntDivideDirective,
        U64Value: UnsignedIntDivideDirective,
        F64Value: FloatDivideDirective,
    },
    BinaryStackOp.FLOOR_DIVIDE: {
        I64Value: SignedIntDivideDirective,
        U64Value: UnsignedIntDivideDirective,
        # special case for float floor div
    },
    BinaryStackOp.GREATER_THAN: {
        I64Value: SignedGreaterThanDirective,
        U64Value: UnsignedGreaterThanDirective,
        F64Value: FloatGreaterThanDirective,
    },
    BinaryStackOp.GREATER_THAN_OR_EQUAL: {
        I64Value: SignedGreaterThanOrEqualDirective,
        U64Value: UnsignedGreaterThanOrEqualDirective,
        F64Value: FloatGreaterThanOrEqualDirective,
    },
    BinaryStackOp.LESS_THAN_OR_EQUAL: {
        I64Value: SignedLessThanOrEqualDirective,
        U64Value: UnsignedLessThanOrEqualDirective,
        F64Value: FloatLessThanOrEqualDirective,
    },
    BinaryStackOp.LESS_THAN: {
        I64Value: SignedLessThanDirective,
        U64Value: UnsignedLessThanDirective,
        F64Value: FloatLessThanDirective,
    },
    BinaryStackOp.EQUAL: {
        I64Value: IntEqualDirective,
        U64Value: IntEqualDirective,
        F64Value: FloatEqualDirective,
    },
    BinaryStackOp.NOT_EQUAL: {
        I64Value: IntNotEqualDirective,
        U64Value: IntNotEqualDirective,
        F64Value: FloatNotEqualDirective,
    },
    BinaryStackOp.OR: {BoolValue: OrDirective},
    BinaryStackOp.AND: {BoolValue: AndDirective},
}
